//
//  Guit GUI Toolkit Examples
//  Copyright © 2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <guit.hpp>
#include <gshape.hpp>
#include "examples.hpp"
using namespace std;
using namespace guit;

class DiagramCanvas : public GChoiceBox {
public:
  enum Action {Select, InitState, State, Transition};

  DiagramCanvas();
  void erase();
  void eraseAll();
  void restore();
  Gadget* findNode(GPoint const& wpos, Gadget* except);

  GBool changed;
  GChoice chosen_action;
  GText info;

protected:
  float minsize_{5.f};
  GBox undolist_;
  GListBox statecontrols_;
  void canvasMouseDown(GMouseEvent&);
  void linkMouseUp(GMouseEvent&);
  void attachControls(Gadget* shape);
};


class Diagram : public GWindow {
  DiagramCanvas canvas;
  GText rmode, connectmsg;

public:
  enum RemoteMode {StandAlone, Client, Server};
  
  Diagram() : GWindow("#diagram Diagram Editor") {

    auto& opendial = OpenDialog("Open File",":/data")
    // shows an alert before opening the dialog if canvas.changed is true
    .alertIf(canvas.changed);

    // reads gadgets and add them to the canvas
    opendial << [&] {
      canvas.read(opendial.pathname(), GIOMode::Content);
      canvas.changed = false;
    };

    auto& savedial = SaveDialog("Save File","~");
    savedial << [&] {
      canvas.write(savedial.pathname(), GIOMode::Content);
      canvas.changed = false;
    };

    auto& quitdial = QuitDialog("#quit-dial Close Window",
                                "Do you want to close the window?")
    // open QuitDialog only if canvas.changed is true, otherwise just quit
    .openOnlyIf(canvas.changed);


    auto& toolbar = ToolBar("#toolbar")
    << (".open-btn" << opendial)
    << (".save-btn" << savedial)
    << (ListBox("#actions")
        << canvas.chosen_action
        << ".select-btn"
        << ".initstate-btn"
        << ".state-btn"
        << ".transition-btn"
        )
    << (".untrash-btn"  << [&]{canvas.restore();})
    << (".trash-btn"    << [&]{canvas.erase();})
    << (".trashall-btn" << [&]{canvas.eraseAll();})
    << (".print-btn"    << [&]{canvas.write(cout, GIOMode::Content);});

    *this
    << quitdial
    << toolbar
    << (ScrollBox << canvas)
    << (StatusBar << (Label << canvas.info));

    canvas.changed = false;
  }
};


GWindow& Examples::makeDiagram() {
  return *new Diagram;
}


DiagramCanvas::DiagramCanvas() :
GChoiceBox("#canvas"),
statecontrols_("{flow:row; align:c c}")
{
  *this
  << on::mousedown / [this](GEvent& e){canvasMouseDown(e.asMouseEvent());}
  << on::change / [this]{attachControls(chosen());};

  statecontrols_
  << chosen_action
  << ".select-btn"
  << ".initstate-btn"
  << ".state-btn"
  << ".transition-btn";

  info = "Ready";
  chosen_action = Action::Select;   // selection is initial mode

  // when not in TSelect (selection) mode, the canvas catches events
  // which allows drawing on the top of an existing shape
  chosen_action << [&]{catchEvents(chosen_action != Action::Select);};
}


void DiagramCanvas::attachControls(Gadget* shape) {
  if (shape && !shape->to<GLineShape>()) {
    statecontrols_ << new GRelativePos(*shape, GPoint(0,-30));
    *this << statecontrols_;
  }
  else remove(statecontrols_);

}


void DiagramCanvas::canvasMouseDown(GMouseEvent& e) {
  info = "";
  choose(nullptr);

  switch (chosen_action) {

    case Action::InitState:
      *this
      << (ArcShape(".initstate") << new GDragPos(e.pos()));
      changed = true;
      chosen_action = Action::Select;
      break;

    case Action::State: {
      *this
      << (VBox(".state")
          << new GDragPos(e.pos())
          << new GDragSize()
          << "Name:"
          << TextField
          << "Action:"
          << TextField
          ).setCheckable(true).setDownActivable(true);
      changed = true;
      chosen_action = Action::Select;
    } break;


    case Action::Transition: {
      // find node on which the link was pressed in the canvas
      auto* from = findNode(e.wpos(), nullptr);
      if (!from) {
        info = "Link must start from a node!";
      }
      else {
        auto& link = LineShape(".transition")
        // A REVOIR: l'ordre importe et le second appelle handleUp
        // sur de la memoire detruite !
        << on::mouseup / [this](GEvent& e){linkMouseUp(e.asMouseEvent());}
        << new GDragSize();

        *this << link;
        link.changePosProp(*new GRelativePos(*from, e.wpos() - from->wpos()));
        link.reshapeInteractively(e);
      }
    } break;

  }
}


void DiagramCanvas::linkMouseUp(GMouseEvent& e) {
  auto* link = e.from().to<GLineShape>();
  if (!link) return;
  auto& wpos = e.wpos();
  GHandle loc = link->findHandle(wpos);

  // find gadget (except itself!) on which the link was released
  auto* target = findNode(wpos, this);

  switch (loc) {
    case GHandle::Point1:
      if (!target) {
        info = "Link must start from a node!";
        remove(link);
        changed = true;
      }
      else {
        link->changePosProp(*new GRelativePos(*target, e.wpos() - target->wpos()));
        changed = true;
        choose(nullptr);
      }
      break;

    case GHandle::Point2:
      if (!target) {
        info = "Link must end on a node!";
        remove(link);
        changed = true;
      }
      else {
        link->changePos2Prop(*new GRelativePos(*target, e.wpos() - target->wpos()));
        changed = true;
        choose(nullptr);
      }
      break;

    default:
      return;
  }
  chosen_action = DiagramCanvas::Select;
}


Gadget* DiagramCanvas::findNode(GPoint const& wpos, Gadget* except) {
  // find gadget on which the link was released
  // ignore 'execpt' and GLineShape shapes
  return childAtWPoint(wpos, false, [except](Gadget& shape){
    return &shape != except && !shape.to<GLineShape>();
  });
}


void DiagramCanvas::erase() {
  if (auto* item = chosen()) {
    undolist_ << *item;    // so that we can restore the item later
    choose(nullptr);
    if (childCount() == 0) changed = false;
    update();
  }
}


void DiagramCanvas::eraseAll() {
  choose(nullptr);
  auto list = children();  // copy needed because the loop modifies chidren
  for (auto& item : list) undolist_ << *item;
  changed = false;
  update();
}


void DiagramCanvas::restore() {
  if (auto* item = undolist_.childAt(-1)) *this << *item;
}

