//
//  OpenGL Example
//  Guit GUI Toolkit
//  Copyright © 2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#if defined(_WIN32)
#  include <windows.h>
#endif
#if defined(__APPLE__)
#  define GL_SILENCE_DEPRECATION
#  include <OpenGL/gl.h>
#else
#  include <GL/gl.h>
#endif

#include <guit.hpp>
#include <ggraphics.hpp>
#include "examples.hpp"
using namespace std;
using namespace guit;


class Canvas : public GBox {
public:
  Canvas();
  GFloat red{0, 0, 1}, green{0, 0, 1}, blue{0, 0, 1};
  GFloat xrot{0, 0, 1}, yrot{0, 0, 1}, zrot{0, 0, 1};
  GFloat xtrans{0, -1, 1}, ytrans{0, -1, 1};
  GTimer timer{50};
  GItem infos;

private:
  void paint(GEvent&);
  void mouseDown(GMouseEvent&);
  void mouseUp(GMouseEvent&);
  void mouseDrag(GMouseEvent&);
  void showInfos(GMouseEvent&);
  void timerCB();
};


class OpenGL : public GWindow {
  Canvas canvas;
  GCheck runanim{};  // prop that start/stops animation

public:
  OpenGL() : GWindow("#opengl OpenGL")
  {
    // start/stop animation depending on the state of runanim
    runanim << [this]{canvas.timer.start(runanim());};

    // controls whether the 'controls' box is shown
    auto& show_controls = Check(true);

    // menubar and pulldown menus
    auto& menubar = MenuBar()
    << ("File"
        << (Menu(".menu") << ("Quit" << []{GApp::close();}))
        )
    << ("Colors"
        << (Menu(".menu .color_menu")
            << createSlider("R", canvas.red)
            << createSlider("G", canvas.green)
            << createSlider("B", canvas.blue)
            )
        )
    << ("Animation"
        << (Menu(".menu")
            << (Button()
                << runanim    // value bound with this prop
                << on::checked   / "  Stop  "   // text when selected
                << on::unchecked / "  Start  "  // text when deselected
                )
            )
        )
    << Filler()
    << (ChoiceItem()         //  shows/hides the controls box
        << show_controls
        << on::checked   / "<b> \\< "
        << on::unchecked / "<b> > "
        );
    
    
    // contains gadgets that control the colors etc.
    auto& controls =
    VBox("#controls")
    << (Show() <<= show_controls)   // show when show_controls is true
    << (HBox(".panel Animation")
        << Filler()
        << (Button()
            << runanim
            << on::checked   / "  Stop  "
            << on::unchecked / "  Start  "
            )
        )
    << (VBox(".panel Color")
        << createSlider("R", canvas.red)
        << createSlider("G", canvas.green)
        << createSlider("B", canvas.blue)
        )
    << (VBox(".panel Rotation")
        << createSlider("X", canvas.xrot)
        << createSlider("Y", canvas.yrot)
        << createSlider("Z", canvas.zrot)
        )
    << (VBox(".panel Translation")
        << createSlider("X", canvas.xtrans)
        << createSlider("Y", canvas.ytrans)
        );

    *this
    << on::show / [&]{runanim = true;}
    << on::hide / [&]{runanim = false;}
    << menubar
    << GFlow::stack
    << canvas
    << (HBox("{align: e x; background:none}") << controls);

    canvas.red = 1.f;
  }
  
  Gadget& createSlider(GString const& label, GFloat& val) {
    return HBox(".textslider")
    << label
    << (Slider << val)
    << (TextField << val);
  }
};


GWindow& Examples::makeOpenGL() {
  return *new OpenGL;
}


Canvas::Canvas() : GBox("#canvas") {
  // update view when the values of red, green, blue, etc. have changed
  red    << [this]{update();};
  green  << [this]{update();};
  blue   << [this]{update();};
  xrot   << [this]{update();};
  yrot   << [this]{update();};
  zrot   << [this]{update();};
  xtrans << [this]{update();};
  ytrans << [this]{update();};
  timer  << [this]{timerCB();};

  infos << "infos" << GColor::yellow << GFont::large;
  infos.show(false);

  *this
  << infos
  << GBackground::black
  << on::paint     / [this](auto& e){paint(e);}
  << on::mousedown / [this](auto& e){mouseDown(e.asMouseEvent());}
  << on::mouseup   / [this](auto& e){mouseUp(e.asMouseEvent());}
  << on::mousedrag / [this](auto& e){mouseDrag(e.asMouseEvent());};
}

void Canvas::timerCB() {
  xrot = xrot() + 0.005f;
  if (xrot() >= 1.f) xrot = 0.f;

  yrot = yrot() + 0.005f;
  if (yrot() >= 1.f) yrot = 0.f;
}

void Canvas::mouseDown(GMouseEvent& e) {
  infos << new GPos(e.pos());
  infos.show(true);
  showInfos(e);
}

void Canvas::mouseUp(GMouseEvent& e) {
  showInfos(e);
  infos.show(false);
}

void Canvas::mouseDrag(GMouseEvent& e) {
  showInfos(e);
}

void Canvas::showInfos(GMouseEvent& e) {
  infos << std::to_string(e.x()) + "," + std::to_string(e.y());
}

void Canvas::paint(GEvent& e) {
  GGraphics g(e);
  g.beginGL();
  int side = std::min(int(g.width()), int(g.height()));

  /*
   glBegin(GL_LINE_STRIP);
   glColor4f(1, 1, 0, 1);
   glVertex2f(globpos().x, globpos().y);
   glVertex2f(globpos().x+g.width(), globpos().y+g.height());
   glEnd();

   glViewport(g.glx(), g.gly(), side, side);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(-2, +2, -2, +2, 0, 30);
   glMatrixMode(GL_MODELVIEW);

   glBegin(GL_LINE_STRIP);
   glColor4f(1, 0, 0, 1);
   glVertex2f(-2, -2);
   glVertex2f(2, 2);
   glEnd();
   */

  glPushAttrib(GL_ALL_ATTRIB_BITS);
  glPushMatrix();

  glViewport(GLint(g.glx()), GLint(g.gly()), side, side);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-2, +2, -2, +2, 1.0, 15.0);
  glMatrixMode(GL_MODELVIEW);

  //glClear(GL_COLOR_BUFFER_BIT) clear GUI space !!!
  //glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glClear(GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);

  GLfloat lightPosition[4] = {0, 0, 10, 1};
  glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);

  GLfloat lightColor[4] = {red, green, blue, 1};
  glLightfv(GL_LIGHT0, GL_AMBIENT, lightColor);

  glLoadIdentity();
  //glTranslatef(xtrans(), ytrans(), -10 - 10*ztrans());
  glTranslatef(xtrans, ytrans, -10);
  glRotatef(xrot*360, 1.0, 0.0, 0.0);
  glRotatef(yrot*360, 0.0, 1.0, 0.0);
  glRotatef(zrot*360, 0.0, 0.0, 1.0);

  glBegin(GL_QUADS);
  glNormal3f(0,0,-1);
  glVertex3f(-1,-1,0);
  glVertex3f(-1,1,0);
  glVertex3f(1,1,0);
  glVertex3f(1,-1,0);
  glEnd();

  glBegin(GL_TRIANGLES);
  glNormal3f(0.f, -1.f, 0.707f);
  glColor3f(1.0f, 0.0f, 0.0f);
  glVertex3f(-1,-1,0);
  glColor3f(1.0f, 1.0f, 0.0f);
  glVertex3f(1,-1,0);
  glColor3f(1.0f, 0.0f, 1.0f);
  glVertex3f(0,0, 1.2f);
  glEnd();

  glBegin(GL_TRIANGLES);
  glNormal3f(1, 0, 0.707f);
  glVertex3f(1,-1,0);
  glVertex3f(1,1,0);
  glVertex3f(0,0, 1.2f);
  glEnd();

  glBegin(GL_TRIANGLES);
  glNormal3f(0, 1, 0.707f);
  glVertex3f(1, 1, 0);
  glVertex3f(-1, 1, 0);
  glVertex3f(0, 0, 1.2f);
  glEnd();

  glBegin(GL_TRIANGLES);
  glNormal3f(-1, 0, 0.707f);
  glVertex3f(-1, 1, 0);
  glVertex3f(-1, -1, 0);
  glVertex3f(0, 0, 1.2f);
  glEnd();

  glPopAttrib();
  glPopMatrix();
  g.endGL();
}
