//
// Gadget Rendering
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
#include <algorithm>
#include <gapp.hpp>
#include <gevent.hpp>
#include <gtrigger.hpp>
#include <ggraphics.hpp>
#include <gtext.hpp>
#include <gicon.hpp>
#include <gcolor.hpp>
#include <gfont.hpp>
#include <gformat.hpp>
#include <gborder.hpp>
#include <gtextitem.hpp>
#include <gbox.hpp>
#include <gwindow.hpp>
#include <gpos.hpp>
#include <gsize.hpp>
//#include <ginspector.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gtextrender.hpp>
#include <gstyle.hpp>
#include <gadgetstyle.hpp>
#include <gnative.hpp>
using namespace std;
namespace guit {

// SDL defines main as SDL_main, which generates an error with align.main below
#undef main

GRenderGraphics::GRenderGraphics(GWindow& win) : GGraphics(win),
area_(0, 0, win.frame_.w, win.frame_.h),
cliprect_(area_) {
  scale_ = win.scale_;
}

bool GRenderGraphics::setClipRect(Gadget& obj) {
  GClip clip(cliprect_, area_);
  clip.intersection(obj.frame_);
  if (clip.empty()) return false;

  cliprect_ = {clip.x1, clip.y1, clip.x2-clip.x1, clip.y2-clip.y1};
  setWinClip(cliprect_);
  return true;
}

GRender& GRender::operator=(GRender const& from) {
  parr_ = nullptr;
  level_ = 0;
  instyles_ = nullptr;
  child_inscount_ = own_inscount_ = 0;
  obj_      = from.obj_;
  text_     = from.text_;
  hotkey_   = from.hotkey_;
  textpos_  = from.textpos_;
  textsize_ = from.textsize_;
  font_     = from.font_;
  flow_     = from.flow_;
  pos_      = from.pos_;
  size_     = from.size_;
  flex_     = from.flex_;
  format_   = from.format_;
  color_    = from.color_;
  background_ = from.background_;
  icon_     = from.icon_;
  align_    = from.align_;
  margin_   = from.margin_;
  border_   = from.border_;
  memcpy(specif, from.specif, sizeof(specif));
  return *this;
}

GRender::~GRender() {
  if (level_ == 0) delete instyles_;
}


GRender::GRender(Gadget& obj, GRender* parr) :
obj_(&obj), parr_(parr)
{
  if (parr) {
    level_ = parr->level_ + 1;
    instyles_ = parr->instyles_;
    child_inscount_ = own_inscount_ = parr->child_inscount_;

    // inherited props                                       hardcoded!!!!!!
    font_ = parr->font_;
    color_ = parr->color_;
    background_ = parr->background_;
    obj.settings_.trulyEnabled
    = obj.modes_.enabled && parr->obj_->settings_.trulyEnabled;
  }
  else {
    level_ = 0;
    instyles_ = new std::vector<StyleNode>(100);
    child_inscount_ = own_inscount_ = 0;
    font_ = GFont::regular.value();
    color_ = &GColor::black.value();
    obj.settings_.trulyEnabled = obj.modes_.enabled;

    /*
    if (GCore::mainWindow() && &obj != GCore::mainWindow()) {
      if (GadgetStyle* gs = GCore::mainWindow()->style()) {
        for (auto* it = gs->ids_; *it != 0; ++it) {  // 0 ended
          cerr <<"super " << *it << endl;
          if (GStyle* s = GStyleBase::instance().findStyle(*it)) addCombinator(s);
        }
      }
    }*/
    /*
    if (auto* ps = obj.getPseudoParent()) {
      if (auto* win = ps->getWindow()) {
        if (GadgetStyle* gs = win->style()) {
          auto* t = win->findChild<GTitle>();
          cerr <<"super 2" << (t ? t->text() : "non") << endl;
          for (auto* it = gs->ids_; *it != 0; ++it) {  // 0 ended
            if (GStyle* s = GStyleBase::instance().findStyle(*it)) addCombinator(s);
          }
        }
      }
    }*/
  }

  auto& class_style = obj.type().style();

  applyStyles(&obj, class_style);                    // class & instance styles

  if (instyles_) {                                          // inherited styles
    auto end = instyles_->begin() + own_inscount_;
    for (auto it = instyles_->begin(); it < end; ++it) {
      if (it->style_->has_desc_) applyInStyles(*it, &obj, class_style);
    }
  }
                                        // props and inline styles of the gadget
  if (!obj_->props_.empty()) {
    for (auto& p : obj_->props_) p->apply(obj_, *this, GStyle::InlineSpecif);
  }

  // font and colors must be defined!
  if (!font_.impl) {
    gerror("GRender","internal error: null font");
    font_ = GFont::regular.value();  // initialisee ?
  }
  if (!color_) {
    gerror("GRender","internal error: null color");
    color_ = &GColor::black.value();
  }

  if (border_ && border_->hasPickShape(*obj_)) {
    obj_->setShape(border_);
  }
}


void GRender::addCombinator(GStyle* s) {
  if (instyles_->size() <= child_inscount_) {
    instyles_->resize(child_inscount_ + 50);
  }
  (*instyles_)[child_inscount_++] = StyleNode(level_, s);
}


void GRender::applyStyles(Gadget* g, GStyle& class_style) {
  auto& base = GStyleBase::instance();
  bool isbox = g->toBox() != nullptr;

  // style of class
  class_style.apply(g, *this);                             // style applies to g
  if (isbox && class_style.has_desc_) addCombinator(&class_style); // style applies to desc

  // styles of instance
  if (GadgetStyle* inst_styles = g->style_) {
    if (!inst_styles->realized_) inst_styles->realize(*g);

    if (inst_styles->ids_) {
      for (GId* it = inst_styles->ids_; *it != 0; ++it) {  // 0 terminated table
        if (GStyle* s = base.findStyle(*it)) {
          s->apply(g, *this);                              // style applies to g
          if (isbox && s->has_desc_) addCombinator(s);     // style applies to desc
        }
      }
    }
  }
}

#define APPLY_DESC(styles, g, id) { \
  auto it2 = styles->find(id); \
  if (it2 != styles->end()) { \
    it2->second->apply(g, *this); \
    if (it2->second->has_subrule_) addCombinator(it2->second); \
  } \
}


// inherited styles
void GRender::applyInStyles(StyleNode& asc_style, Gadget* g, GStyle& class_style)
{
  GStyleMap* desc_styles = asc_style.style_->stylemap_;
  if (!desc_styles) return; // should not occur

  int asc_level = asc_style.level_;
  bool any_desc = asc_style.style_->any_desc_;
  bool direct_desc = asc_style.style_->direct_desc_;

  // style of class: rule for any child
  if (any_desc) {
    APPLY_DESC(desc_styles, g, class_style.id_);
  }
  // style of class: rule for direct child (which ID is normal ID + 1)
  if (direct_desc && level_ == asc_level + 1) {
    APPLY_DESC(desc_styles, g, class_style.id_ + GStyle::DirectChild);
  }
  
  // styles of instance
  if (GadgetStyle* inst_styles = g->style_) {
    if (!inst_styles->realized_) inst_styles->realize(*g);
    if (inst_styles->ids_) {
      for (GId* it = inst_styles->ids_; *it != 0; ++it) {
        // rules for any desc
        if (any_desc) {
          APPLY_DESC(desc_styles, g, *it);
        }
        // rules for direct desc
        if (direct_desc && level_ == asc_level + 1) {
          APPLY_DESC(desc_styles, g, *it + GStyle::DirectChild);
        }
      }
    }
  }
}


void GRender::print(std::ostream& out) {
  GString flow, font, align, gap;
  gconvert(flow_, flow);
  gconvert(font_, font);
  gconvert(align_, align);
  gconvert(gap_, gap);
  out
  << "Type: "   << obj_->typeName() <<  "\n"
  << "Text: "    << (text_ ? *text_ : "")<< "\n"
  << "Hotkey: "  << (hotkey_ ? *hotkey_ : "")<< "\n"
  << "Flow: "    << flow << "\n"
  << "Pos: "     << textpos_ << "\n"
  << "Size: "    << size_ << "\n"
  << "TextSize: "<< pos_ << "\n"
  << "Font: "    << font << "\n"
  << "Format: "  << format_ << "\n";
  out << "Color: ";
  if (color_) out << *color_;
  out << "\n";
  out << "Background: ";
  if (background_) out << *background_;
  out << "\n";
  out << "Icon: " << icon_ << "\n";
  out << "Align: " << align << "\n";
  out << "Gap: " << gap << "\n";
  out << "Margin: ";
  if (margin_) {
    GString margin;
    gconvert(*margin_, margin);
    out << margin;
  }
  out << "\n";
  out << "Border: ";
  if (border_) out << *border_;
  out << "\n"<< endl;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

struct TopCh {
  decltype(GadgetLayout::layer) layer;
  Gadget* ch;
  static bool comp(TopCh const& a, TopCh const& b) {return a.layer < b.layer;}
};

void GRender::paintBox(GRenderGraphics& g, GBox& box) {
  if (!box.isShown() || box.frame_.w <= 0.f || box.frame_.h <= 0.f) return;

  GRect par_area = g.area_;
  GRect par_clip = g.cliprect_;

  auto* blist = box.bchildren_;
  paintGadget(g, blist ? (DrawBackground|DrawBorder) : DrawAll);
  
  GRect ch_area = g.area_;
  GRect ch_clip = g.cliprect_;

  if (blist) {
    float delta_w{}, delta_h{};
    
    for (auto& it : *blist) {
      g.area_ = ch_area;
      g.cliprect_ = ch_clip;
      it.child_->doPaint(g, this);
      
      if (it.place_ == GBox::Place::Right)    // !!!!!! IL EN MANQUE !!!!
        delta_w = it.child_->w();
      else if (it.place_ == GBox::Place::Bottom)
        delta_h = it.child_->h();
    }
    // +2 needed because setClip() is rounded
    ch_area.w -= delta_w +2;
    ch_area.h -= delta_h +2;
    ch_clip.w -= delta_w +2;
    ch_clip.h -= delta_h +2;
  }
  
  // the box contains gadgets with a level > 0 that must be treated first
  //TopCh topch[box.children_.size()];
  std::vector<TopCh> topch(box.children_.size());
  size_t topcount = 0;

  for (auto& ch : box.children_) {
    if (ch->isShown() && !ch->isTopLevel()) {
      if (ch->layout_.layer == 0) {
        g.area_ = ch_area;
        g.cliprect_ = ch_clip;
        ch->doPaint(g, this);
      }
      else topch[topcount++] = {ch->layout_.layer, ch};
    }
  }

  // specialclip
  if (obj_->modes_.unclippedLayers && topcount > 0) {
    ch_area = par_area;
    ch_clip = par_clip;
  }

  if (topcount == 0) {
    box.layout_.hasLayers = 0;  // used by pick()
  }
  else {
    box.layout_.hasLayers = 1;
    
    if (topcount == 1) {
      g.area_ = ch_area;
      g.cliprect_ = ch_clip;
      topch[0].ch->doPaint(g, this);
    }
    else {
      // stable_sort preserves order
      //std::stable_sort(topch, topch+topcount, TopCh::comp);
      std::stable_sort(topch.begin(), topch.begin() + topcount, TopCh::comp);
      for (size_t k = 0; k < topcount; ++k) {
        g.area_ = ch_area;
        g.cliprect_ = ch_clip;
        topch[k].ch->doPaint(g, this);
      }
    }
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GRender::paintPieBackground(GRenderGraphics& gr) {
  //gr.setThickness(0);
  gr.drawPie(gr.area_, obj_->angleStart(), obj_->angleExtent());

  if (obj_->isChecked()) {
    setAutoColor(gr);
    gr.unsetBgColor();
    gr.setThickness(2);
    gr.drawArc(GRect(gr.area_.x+5, gr.area_.y+5, gr.area_.w-15, gr.area_.h-15),
               obj_->angleStart() + 5, obj_->angleExtent() - 10);
  }
}

bool GRender::paintGadget(GRenderGraphics& g, int options) {
  if (!obj_->isShown() || obj_->frame_.w <= 0.f || obj_->frame_.h <= 0.f) return false;
  if (!g.setClipRect(*obj_)) return false;
  
  g.area_ = obj_->frame_;
  textpos_  = {0.f, 0.f};
  textsize_ = {0.f, 0.f};

  bool pieitem{};
  auto* parent = parr_ ? parr_->obj_ : nullptr;
  if (parent && parent->isPie()) pieitem = true;

  // draw background
  if (pieitem) {
    g.unsetColor();
    g.setBgColor(background_);
    paintPieBackground(g);
  }
  else if (background_
           // && border_  // normally border_ is never null
           && !border_->hasDrawShape(*obj_)
           && (options & DrawBackground)) {
    g.unsetColor();
    g.setBgColor(background_);
    g.drawRect(g.area_);
  }

  if (margin_ && (options & AddMargin)) {
    g.area_.x += margin_->left;
    g.area_.y += margin_->top;
    g.area_.w -= margin_->left + margin_->right;
    g.area_.h -= margin_->top  + margin_->bottom;
  }
  
  // draw border and background inside it
  if (!pieitem && (options & DrawBorder)) {
    // color & bgcolor must be set otherwise the previous colors will be used!
    if (obj_->isActive() || obj_->isChecked()) g.setColor(border_->activecolor);
    else g.setColor(border_->idlecolor);

    if (options & DrawBackground) g.setBgColor(background_);
    else g.unsetBgColor();
    border_->drawBorder(g, *this);

    // dont add thickness, it should be inside border space
    g.area_.x += border_->left;
    g.area_.y += border_->top;
    g.area_.w -= border_->left + border_->right;
    g.area_.h -= border_->top  + border_->bottom;
  }

  if (GCore::inspecting() && !obj_->isInternal() && !g.window()->isInternal()) {
    if (obj_ == GCore::inspected()) g.setColor(GCore::inspectedColor());
    else g.setColor(GCore::inspectColor());
    g.setBgColor((GRgba*)nullptr);
    g.drawRect(obj_->frame_);
  }

  if ((options & DrawText) && text_ && !obj_->toBox()) {
    drawText(g, -1);
  }

  // must be after drawText() because it sets textpos_ and textsize_
  if (icon_) {
    auto* w = obj_->toWindow();
    if (!w) drawIcon(g);   // the icon of the window is special
  }

  if (hotkey_) drawHotkey(g);

  // called *before* drawing the children but *after* the background
  if ((obj_->eventmask_ & GEventType::Paint) != 0) {
    GEvent e(on::paint, GCore::elapsedTime());
    e.window_ = g.win_;
    e.gadget_ = obj_;
    obj_->doFire(e);
  }
  return true;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GRender::drawText(GRenderGraphics& g, float offset) {
  GString const* text = text_;
  GString formatted_text;  // used when there is a text format

  if (format_ && !text->empty()) {
    if (formatText(*format_, *text_, formatted_text)) text = &formatted_text;
  }

  auto* tf = obj_->toTextItem();
  if (tf) {       // nb: call even if text is empty to draw caret
    g.cliprect_.w -= 2;   // text must not be draw on the border!
    g.cliprect_.h -= 2;
  }
  else if (text->empty()) return;

  auto* attrlist = tf ? tf->attrlist_ : nullptr;
  auto area = g.area_;
  bool richtext = obj_->modes_.richText;
  bool ispie = parr_ && parr_->obj_->isPie();

  ssize_t caret{-1};
  size_t selbegin{}, selend{};

  if (tf && GApp::currentFocus() == obj_) {
    caret = tf->ctrl_.caret - tf->ctrl_.offset;
    selbegin = size_t(tf->ctrl_.selbegin - tf->ctrl_.offset);
    selend = size_t(tf->ctrl_.selend - tf->ctrl_.offset);
  }

  GItem* item{};
  if (obj_->isMultiLine() && (item = obj_->toItem())) {
    auto* cache = item->tcache_;
    if (!cache) {
      gerror("drawText","internal error: no text cache!");
      goto END;
    }

    //                      !!!! refaire seulement qd on change le layout !!!

    GTextRender::layoutText(font_, richtext, *text, area.w, attrlist, cache);

    GTextRender tr(font_, richtext, color_, selbegin, selend, attrlist);
    float x = area.x, y = area.y;

    for (auto& it : *cache->rows_) {
      tr.draw(g, x, y+it.h_, text->substr(it.start_,it.length_), it.start_);

      if (caret >= ssize_t(it.start_) && caret <= ssize_t(it.start_ + it.length_)) {
        drawCaret(g, *text, font_, attrlist, it.start_, caret,
                  richtext, x, y, it.h_);
      }
      y += it.h_;
    }
  }

  else {   // not multines
    GString buf;

    if (tf && tf->ctrl_.offset > 0 && size_t(tf->ctrl_.offset) <= text->size()) {
      buf.assign(*text, size_t(tf->ctrl_.offset));
      text = &buf;
      //text.erase(0, tf->offset_);
    }

    textsize_ = font_.textSize(*text, richtext, attrlist);

    float iconw = textsize_.h + gap_.mainspacing;

    if (ispie) {
      float cx = area.x + area.w /2.f;
      float cy = area.y + area.h /2.f;
      float a = -(obj_->angleStart() + obj_->angleExtent() / 2.f) * float(M_PI / 180);
      textpos_.x = cx + area.w / 3.5f * std::cos(a) - textsize_.w /2.f;
      textpos_.y = cy + area.h / 3.5f * std::sin(a) - textsize_.h;
      if (offset >= 0) setTextAutoColor(g);
    }
    else {
      // align text in menus.
      bool leftalign = parr_ && parr_->obj_->modes_.menuLayout
      && !parr_->obj_->isRow();

      // error C2039: 'SDL_main' n'est pas membre de 'guit::GAlignValue'  $$$$$
      switch (align_.main) {
        case GAlign::Start:
          textpos_.x = area.x;
          if (leftalign) textpos_.x += GPrefs::iconWidth();
          else if (icon_ && align_.icon == GAlign::Start) textpos_.x += iconw;
          break;

        case GAlign::End:
          textpos_.x = area.x + area.w - textsize_.w;
          if (icon_ && align_.icon == GAlign::End) textpos_.x -= iconw;
          break;

        case GAlign::Center:
        case GAlign::Extend:
          if (leftalign)
            textpos_.x = area.x + GPrefs::iconWidth()/2.f
            + (area.w - textsize_.w) /2.f;
          else if (!icon_)
            textpos_.x = area.x + (area.w - textsize_.w) /2.f;
          else if (align_.icon == GAlign::End)
            textpos_.x = area.x + (area.w - textsize_.w - iconw) /2.f;
          else
            textpos_.x = area.x + iconw/2.f + (area.w - textsize_.w) /2.f;
          break;
      }

      switch (align_.cross) {
        case GAlign::Start:
          textpos_.y = area.y;
          break;

        case GAlign::End:
          textpos_.y = area.y + (area.h - textsize_.h);
          break;

        case GAlign::Center:
        case GAlign::Extend:
          textpos_.y = area.y + (area.h - textsize_.h) / 2.f;
          break;
      }

      if (offset >= 0) {
        if (auto* parent = parr_ ? parr_->obj_ : nullptr) {
          if (parent->isRow())
            setTextAutoColor(g, textpos_.y, offset);
          else
            setTextAutoColor(g, textpos_.x, offset);
          }
        }
      }

    GTextRender tr(font_, richtext, color_, selbegin, selend, attrlist);
    tr.draw(g, textpos_.x, textpos_.y+textsize_.h, *text, 0);

    if (caret >= 0) {
      drawCaret(g, *text, font_, attrlist, 0, caret, richtext,
                textpos_.x, textpos_.y, textsize_.h);
    }
  }

  END:
  if (tf) {
    g.cliprect_.w += 2;
    g.cliprect_.h += 2;
    g.setWinClip(g.cliprect_);
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GRender::setAutoColor(GRenderGraphics& gr) {
  auto& bg = gr.bgcolor_;
  if (bg.alpha <= 0) {
    gr.setColor(GColor::black.value_);
  }
  else {
    float c = (bg.red + bg.blue + bg.green) / 3.f / bg.alpha;
    gr.setColor(c > 0.5f ? GColor::black.value_ : GColor::white.value_);
  }
}

void GRender::setTextAutoColor(GRenderGraphics& gr) {
  auto& bg = gr.bgcolor_;
  if (bg.alpha > 0) {
    float c = (bg.red + bg.blue + bg.green) / 3.f / bg.alpha;
    if (c > 0.5f) color_ =  &GColor::black.value_ ;
    else color_ =  &GColor::white.value_;
  }
}

void GRender::setTextAutoColor(GRenderGraphics& gr, float pos, float index) {
  auto& bg = gr.bgcolor_;
  if (bg.alpha > 0) {
    float c = (bg.red + bg.blue + bg.green) / 3.f * bg.alpha;
    if (c > 0.5f || pos > index) color_ = &GColor::black.value_ ;
    else color_ = &GColor::white.value_;
  }
}

bool GRender::formatText(GFormatValue const& form, GString const& from, GString& to) {
  bool changed = false;

  for (size_t pos = 0; pos < from.size(); ) {
    //ssize_t begin = -1;
    size_t begin{};  // !!

    if (!::isdigit(from[pos])) {
      to.push_back(from[pos]);
      pos++;
      continue;
    }
    else if (from[pos]=='-' || from[pos]=='+' || from[pos]=='.' || from[pos]==',') {
      if (pos+1 < from.size() && ::isdigit(from[pos+1])) begin = pos;
      else {
        to.push_back(from[pos]);
        pos++;
        continue;
      }
    }
    else begin = pos;

    changed = true;

    // integral part
    //ssize_t end = begin;
    size_t end = begin;

    while (end < from.size() && ::isdigit(from[end])) end++;
    pos = end;

    // at least as many chars for the integral part
    if (form.nbchars >= 0 && form.nbchars > ssize_t(end-begin)) {
      to.append(form.nbchars - (end-begin), form.fillchar);  // add fill chars
    }
    to.append(from.substr(begin, end-begin));

    // stop if not . or ,
    if (end >= from.size() || (from[end] != '.' && from[end] != ',')) continue;

    // mantissa
    to.push_back(from[end]);
    end++;  // after . or ,

    begin = end;
    while (end < from.size() && ::isdigit(from[end])) end++;
    pos = end;

    if (form.precision >= 0 && form.precision < ssize_t(end-begin)) {
      to.append(from.substr(begin, form.precision));
    }
    else to.append(from.substr(begin, end-begin));

    // remove ending 0 and . if nothing afterwards
    /*
     ssize_t pos = to.size()-1;
     while (pos >= 0) {
     if (to[pos] != '0') break;
     else --pos;
     }
     if (pos == dot) to.erase(pos);
     else to.erase(pos+1);
     */
  }
  return changed;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GRender::drawHotkey(GRenderGraphics& g) {
  if (!hotkey_ || hotkey_->empty()) return;
  string hotk;

  if (parr_ && parr_->obj_->modes_.menuLayout) hotk = *hotkey_;
  else hotk = " ("+ *hotkey_ +")";

  GDim dim = font_.textSize(hotk, false, nullptr);
  float x = g.area_.x + g.area_.w - dim.w;
  g.drawString(x, textpos_.y+textsize_.h, hotk);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GRender::drawCaret(GRenderGraphics& g, GString const& text,
                        GFontValue const& font, GTextAttributes const* attrs,
                        size_t begline, ssize_t caretpos, bool richtext,
                        float x, float y, float h) {
  GDim dim = font_.textSize(text.substr(begline, caretpos-begline),
                            richtext, attrs);
  g.setColor(GPrefs::caretColor());
  g.setThickness(GPrefs::caretWidth());
  g.drawLine(x+dim.w, y-1.f, x+dim.w, y+h);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void GRender::drawIcon(GRenderGraphics& g) {
  if (!icon_) return;
  float x = 0.f, hsize = 0.f, vsize = 0.f;
  float y = textpos_.y;
  
  if (textsize_.h == 0.f) {  // if there was no text
    textsize_.h = font_.fontHeight();
    y = g.area_.y;
  }
  
  switch (align_.icon) {
    case GAlign::Start:
      vsize = hsize = textsize_.h;
      x = g.area_.x;
      y = g.area_.y + (g.area_.h - vsize) / 2.f;
      break;

    case GAlign::End:
      vsize = hsize = textsize_.h;
      x = g.area_.x + g.area_.w - hsize;
      y = g.area_.y + (g.area_.h - vsize) / 2.f;
      break;

    case GAlign::Center:
      vsize = hsize = textsize_.h;
      x = g.area_.x + (g.area_.w - hsize) / 2.f;
      y = g.area_.y;
      break;

    case GAlign::Extend:
      vsize = hsize = g.area_.h;
      x = g.area_.x + (g.area_.w - hsize) / 2.f;
      y = g.area_.y;
      break;
  };
  
  if (icon_->drawfun == GIcon::Image || icon_->drawfun == GIcon::XPM) {
    g.setBgColor((GRgba*)nullptr);
    g.setColor((GRgba*)nullptr);
    g.drawImage(x, y, hsize, vsize, *icon_);
  }
  else if (icon_->drawfun) {  // Symbol
    (icon_->drawfun)(g, *this, GRect(x,y,hsize,vsize));
  }
}

}
