//
//  CSS Styles
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <sstream>
#include <guit.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gstyle.hpp>
using namespace std;
namespace guit {

class GPseudoStyle : public GStyle {
public:
  struct PsCond {
    virtual ~PsCond() = default;
    virtual bool verifies(Gadget*, GRender&) = 0;
  };

  PsCond** conds_{};  // null terminated

  GPseudoStyle(PsCond** conds) : GStyle(1), conds_(conds) {
    is_pseudo_ = true;
  }

  static GPseudoStyle* create(GStrings const& pseudos) {
    if (pseudos.empty()) return nullptr;

    auto tab = new PsCond*[pseudos.size() + 1];  // null terminated
    int k = 0;

    for (auto& it : pseudos) {
      if (it.empty()) return wrong(tab);

      else if (it[0] == ':') {
        GString fun, arg;
        size_t from = 0;
        strtoken(it, fun, from, "(");
        strtoken(it, arg, from, ")");
        //cerr << "fun:"<<fun << " arg:"<<arg << endl;
        if (auto* cond = getCond(fun, arg)) tab[k++] = cond;
        else return wrong(tab);
      }

      else if (it[0] == '[') {
        //cerr << "item:"<<it  << endl;
        GString attr, arg;
        size_t from = 1;
        auto delim1 = strtoken(it, attr, from, "]=~|^$*");

        if (delim1 <= 0) return wrong(tab);
        else if (delim1 == ']') {
          if (auto* cond = getAttrCond(attr, 0, "")) tab[k++] = cond;
          else return wrong(tab);
        }
        else {
          if (delim1 != '=') {  //  ~= or |= or ^= or $= or *=
            if (from+1 >= it.size() || it[from] != '=') return wrong(tab);
            from++; // remove = after ^ etc.
          }
          auto delim2 = strtoken(it, arg, from, "]");
          if (delim2 != ']') return wrong(tab);

          if (auto* cond = getAttrCond(attr, char(delim1), arg)) tab[k++] = cond;
          else return wrong(tab);
        }
      }

      else return wrong(tab);
    }

    tab[k] = nullptr;  // null terminated
    return new GPseudoStyle(tab);
  }

  static GPseudoStyle* wrong(PsCond** tab) {
    delete [] tab;
    return nullptr;
  }


  struct CondActive : public PsCond {
    bool verifies(Gadget* g, GRender&) override {return g->modes_.active;}
  };

  struct CondChecked : public PsCond {
    bool verifies(Gadget* g, GRender&) override {return g->modes_.checked;}
  };

  struct CondEnabled : public PsCond {
    bool verifies(Gadget* g, GRender&) override {return g->settings_.trulyEnabled;}
  };

  struct CondDisabled : public PsCond {
    bool verifies(Gadget* g, GRender&) override {return !g->settings_.trulyEnabled;}
  };

  struct CondFocus : public PsCond {
    bool verifies(Gadget* g, GRender&) override {return g->modes_.focus;}
  };

  struct CondHover : public PsCond {
    bool verifies(Gadget* g, GRender&) override {
      //return r.obj_->isHover();                         // TODO!!!
      return false;
    }
  };
  
  struct CondEmpty : public PsCond {
    bool verifies(Gadget* g, GRender&) override {
      GBox* box = g->toBox();
      return !box || box->children().empty();
    }
  };

  struct CondFirstChild : public PsCond {
    bool verifies(Gadget* g, GRender&) override {
      GBox* par = g->getParent();
      return par && par->childAt(0) == g;
    }
  };

  struct CondLastChild : public PsCond {
    bool verifies(Gadget* g, GRender&) override {
      GBox* par = g->getParent();
      return par && par->childAt(-1) == g;
    }
  };

  struct CondOnlyChild : public PsCond {
    bool verifies(Gadget* g, GRender&) override {
      GBox* par = g->getParent();
      return par && par->childCount() == 1;
    }
  };

  struct CondNthChild : public PsCond {
    int index = -1;

    CondNthChild(GString const& arg) {index = ::atoi(arg.data());}

    bool verifies(Gadget* g, GRender&) override {
      GBox* par = g->getParent();
      return par && par->childAt(index) == g;
    }
  };

  static PsCond* getCond(GString const& name, GString const& arg) {
    if (name.empty()) return nullptr;

    switch (strid(name)) {
      case strid(":active"): {
        static auto* c = new CondActive;
        return c;
      }
      case strid(":checked"): {
        static auto* c = new CondChecked;
        return c;
      }
      case strid(":hover"): {
        static auto* c = new CondHover;
        return c;
      }
      case strid(":focus"): {
        static auto* c = new CondFocus;
        return c;
      }
      case strid(":enabled"): {
        static auto* c = new CondEnabled;
        return c;
      }
      case strid(":disabled"): {
        static auto* c = new CondDisabled;
        return c;
      }
      case strid(":empty"): {
        static auto* c = new CondEmpty;
        return c;
      }
      case strid(":first-child"): {
        static auto* c = new CondFirstChild;
        return c;
      }
      case strid(":last-child"): {
        static auto* c = new CondLastChild;
        return c;
      }
      case strid(":nth-child"): {
        return new CondNthChild(arg);
      }
      case strid(":only-child"): {
        static auto* c = new CondOnlyChild;
        return c;
      }
    }
    return nullptr;
  }

  struct AttrCond : public PsCond {
    GPropType& proptype;
    AttrCond(GPropType& cl) : proptype(cl) {}
    bool verifies(Gadget* g, GRender&) override {return g->findProp(proptype);}
  };


  struct AttrCondEqual : public AttrCond {
    GString arg;

    AttrCondEqual(GPropType& c, GString const& arg) : AttrCond(c), arg(arg) {}

    bool verifies(Gadget* g, GRender&) override {
      auto* prop = g->findProp(proptype);
      return prop && prop->stringValue() == arg;
    }
  };

  struct AttrCondContain : public AttrCondEqual {
    bool wholeword;

    AttrCondContain(GPropType& c, GString const& arg, bool wholeword)
    : AttrCondEqual(c, arg), wholeword(wholeword) {}

    bool verifies(Gadget* g, GRender&) override {
      auto* prop = g->findProp(proptype);
      if (!prop) return false;
      if (!wholeword) {
        return prop->stringValue().find(arg) != GString::npos;
      }
      else {
        auto s = prop->stringValue();
        auto pos = s.find(arg);
        if (pos ==  GString::npos) return false;

        if (pos > 0 && !::isspace(s[pos-1])) return false;
        pos += arg.size();
        if (s.size() > pos && !::isspace(s[pos]) && s[pos] != '-') return false;
        return true;
      }
    }
  };

  struct AttrCondStart : public AttrCondContain {

    AttrCondStart(GPropType& c, GString const& arg, bool wholeword)
    : AttrCondContain(c, arg, wholeword) {}

    bool verifies(Gadget* g, GRender&) override {
      auto* prop = g->findProp(proptype);
      if (!prop) return false;
      auto s = prop->stringValue();
      auto alen = arg.size();

      if (s.size() < alen) return false;
      if (wholeword && s.size() > alen && !::isspace(s[alen]) && s[alen] != '-')
        return false;
      return s.compare(0, arg.size(), arg) == 0;
    }
  };

  struct AttrCondEnd : public AttrCondEqual {
    AttrCondEnd(GPropType& c, GString const& arg) : AttrCondEqual(c, arg) {}

    bool verifies(Gadget* g, GRender&) override {
      auto* prop = g->findProp(proptype);
      if (!prop) return false;
      auto s = prop->stringValue();
      auto offset = s.size()-arg.size();
      if (offset < 0) return false;
      else return s.compare(offset, arg.size(), arg) == 0;
    }
  };


  static PsCond* getAttrCond(GString const& name, char delim, GString const& arg) {
    //cerr << "Attribute: " << name << " " << delim << " " << arg << endl;
    GPropType* type = GPropType::findType(name);
    if (!type) {
      gerror("GStyle","Unknown attribute "+name+" in CSS selector");
      return nullptr;
    }

    switch (delim) {
      case 0:
        return new AttrCond(*type);
      case '=':
        return new AttrCondEqual(*type, arg);
        break;
      case '~':
        return new AttrCondContain(*type, arg, true);
        break;
      case '*':
        return new AttrCondContain(*type, arg, false);
        break;
      case '|':
        return new AttrCondStart(*type, arg, true);
        break;
      case '^':
        return new AttrCondStart(*type, arg, false);
      case '$':
        return new AttrCondEnd(*type, arg);
        break;
    }
    return nullptr;
  }
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GStyleMap::~GStyleMap() {
  for (auto& it : *this) delete it.second;
}

GStyle* GStyleMap::findStyle(GId selid) {
  auto it = find(selid);
  if (it != end()) return it->second;
  else return nullptr;
}

GStyle& GStyleMap::obtainStyle(GId selid) {
  auto it = find(selid);
  if (it != end()) return *it->second;
  else {
    auto* s = new GStyle(selid);
    //insert_or_assign(selid, s);
    (*this)[selid] = s;
    return *s;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GStyle::~GStyle() {
  delete stylemap_;
}

GStyle::GStyle(GId id) : id_(id) {}

void GStyle::addProp(GProp* prop) {
  if (!prop) gerror("GStyle::addProp","Null prop argument");
  else {
    // GModes is treated specifically via addModes()
    //props_.insert_or_assign(prop->role(), prop);
    props_[prop->role()] = prop;
    prop->addModes(modes_);
  }
}

void GStyle::removeProp(GProp& prop) {
  auto it = props_.find(prop.role());         // !!! INEXACT !!!
  if (it != props_.end()) props_.erase(it);
}

void GStyle::removeProp(GPropType& t) {
  auto it = props_.find(t.role());         // !!! INEXACT !!!
  if (it != props_.end()) props_.erase(it);
}

GProp* GStyle::findProp(GPropType& t) const {
  auto it = props_.find(t.role());
  return (it != props_.end()) ? it->second : nullptr;
}

void GStyle::addDecl(GString const& decl) {
  if (realized_) decl_.clear(); // inutile?

  if (decl_.empty()) decl_ = decl;
  else {
    if (decl_.back() != ';') decl_ += "; ";
    decl_ += decl;
  }
  if (realized_) realize();
}

void GStyle::addStyleDecl(GString const& decl, GProp::Specif spec, GPseudoStyle* ps) {
    if (!ps) {
    addDecl(decl);
    specif_ = spec;
  }
  else {
    pseudos_.push_front(ps);
    ps->is_pseudo_ = true;
    ps->addDecl(decl);
    ps->specif_ = spec;      // specif ok ?
  }
}

GStyle* GStyle::addStyleDecl2(GString const& decl, GProp::Specif spec, GPseudoStyle* ps,
                              GString const& desc, char reltype, bool is_subrule) {
  has_desc_ = true;
  if (!stylemap_) stylemap_ = new GStyleMap;

  auto desc_id = GCore::obtainId(desc);

  if (reltype != '>') any_desc_ = true;
  else {
    direct_desc_ = true;
    desc_id += GStyle::DirectChild;  // watch!
  }

  auto& s2 = stylemap_->obtainStyle(desc_id);
  s2.has_subrule_ = is_subrule;
  s2.addStyleDecl(decl, spec, ps);
  return &s2;
}

void GStyle::realize() {
  realized_ = true;
  if (decl_.empty()) return;

  GString propspec;
  auto len = decl_.size();
  size_t k = 0;
  int level = 0;

  while (k < len) {
    size_t begin = k;
    while (k < len && (decl_[k]!=';' || level > 0)) {
      if (decl_[k] == '{') level++;
      else if (decl_[k] == '}') level--;
      k++;
    }
    if (level > 0) {
      gerror("GStyle","missing } in CSS style: " + decl_);
      return;
    }
    else if (k >= len) {
      propspec = decl_.substr(begin, len-begin);
    }
    else {
      propspec = decl_.substr(begin, k-begin);
      k++;
    }

    GString name, arg;
    auto count = strsplit(propspec, name, arg, ":");

    if (count == 0) continue;

    else if (count == 2 && !name.empty()) {  // NOTE: could share properties !!
      // prop: value
      if (auto* p = GPropType::createInstance(name, arg, true)) {
        p->setImmutable();
        addProp(p);
      }
      else gerror("GStyle","Unknown property '"+ name+"' in CSS style: " + decl_);
    }

    // count = 1 or name empty => :mystyle => inheriting from mystyle
    else {
      if (arg.empty()) {
        gerror("GStyle","Missing value after or before : in CSS style: "+ decl_);
      }
      else {
        if (GStyle* s = GStyleBase::instance().findStyle(arg)) {
          s->realize();
          modes_ = s->modes_;
          for (auto& it : s->props_) addProp(it.second);
        }
        else gerror("GStyle","Unknown inherited style '"+ arg
                    +"' in CSS style: " + decl_);
      }
    }
  }
  decl_.clear();  // neeeded!
}

void GStyle::applyPseudos(Gadget* g, GRender& r) {
  for (auto* it : pseudos_) {

    if (it->conds_) {
      bool stat = true;

      for (auto* c = it->conds_; *c != nullptr; ++c) {   // null terminated
        stat &= ((*c)->verifies)(g,r);
      }

      if (stat) it->apply(g, r);
    }
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GStyleBase& GStyleBase::instance() {
  static GStyleBase instance_;
  return instance_;
}

bool GStyleBase::loadFile(GString const& filename) {
  GIFile in(filename);
  if (!in) return false;
  return load(in, filename);
}

bool GStyleBase::loadBuffer(GString const& buffer) {
  std::istringstream in(buffer);
  return load(in, "buffer");
}

bool GStyleBase::load(std::istream& in, GString const& streamname) {
  bool status = true;
  streamname_ = streamname;

  //in.imbue(std::locale("C"));
  in.imbue(std::locale("en_US.UTF-8"));         // accept accentuated chars !!!!

  while (in) {
    if (!parseStyle(in)) {status = false; goto END;}
    if (in.eof()) {status = true; goto END;}
    if (in.fail()) {
      gerror("CSS::load", "premature end of file");
      status = false;
      goto END;
    }
  }
END:
  return status;
}

void GStyleBase::error(GString const& msg) {
  gerror(streamname_+" CSS", msg);
}

GStyle& GStyleBase::obtainStyle(GId selid) {
  return styles_.obtainStyle(selid);
}

GStyle* GStyleBase::findStyle(GId selid) {
  return styles_.findStyle(selid);
}

GStyle& GStyleBase::obtainStyle(GString const& sel) {
  return styles_.obtainStyle(GCore::obtainId(sel));
}

GStyle* GStyleBase::findStyle(GString const& sel) {
  if (auto id = GCore::findId(sel))
    return styles_.findStyle(id);
  else
    return nullptr;
}


static void skipSpaces(std::istream& in) {
  //char c = 0, c2 = 0;
  int c = 0, c2 = 0;

  while ((c = in.get()) != EOF) {

    if (c == '/' && (c2 = in.peek()) == '*') {
      in.get();
      while ((c = in.get()) != EOF) {
        if (c == '*' && (c2 = in.peek()) == '/') {
          in.get();
          break;
        }
      }
    }
    else if (!::isspace(c)) {
      in.putback(char(c));
      return;
    }
  }
}

static int getSelector(std::istream& in, GString& sel, GStrings& tab) {
  sel.clear();
  //char c = 0;
  int c = 0;
  bool begin = true;

  while ((c = in.get()) != EOF) {
    //char c2 = 0;
    int c2 = 0;

    if (c == '{') {
      return (int)tab.size();
    }
    else if (::isspace(c)) {
      sel += char(c);
      begin = true;
      skipSpaces(in);
    }
    else if (c == '/' && (c2 = in.peek()) == '*') {   // skip comments
      in.get();
      while ((c = in.get()) != EOF) {
        if (c == '*' && (c2 = in.peek()) == '/') {
          in.get();
          break;
        }
      }
    }
    else {  // not space, not comment
      if (begin) {
        begin = false;
        tab.push_back(GString());
      }
      sel += char(c);
      tab.back() += char(c);
    }
  }

  return 0;
}


static bool getDeclBlock(std::istream& in, GString& tokens) {
  tokens.clear();
  int level = 0;
  //char c = 0;
  int c = 0;

  while ((c = in.get()) != EOF) {
    //char c2 = 0;
    int c2 = 0;

    if (c == '}' && level <= 0) {
      return true;
    }
    else if (c == '/' && (c2 = in.peek()) == '*') {
      in.get();
      while ((c = in.get()) != EOF) {
        if (c == '*' && (c2 = in.peek()) == '/') {
          in.get();
          break;
        }
      }
    }
    else {
      if (c == '{') level++;
      else if (c == '}') level--;
      tokens += char(c);
    }
  }

  return false;
}


static bool isValid(GString const& name) {
  if (name.empty()) return false;
  auto c = name[0];
  return ::isalpha(c) || c == '.' || c == '#' || c == ':' || c == '/';
}


bool GStyleBase::parseSelName(GString const& sel, GString& name,
                              GProp::Specif& spec, GPseudoStyle*& ps) {
  name.clear();
  spec = 0;
  GStrings tab, pseudos;

  if (!isValid(sel)) {
    error("invalid selector: " + sel); // + " before {" + decl + "}");
    return false;
  }

  ssize_t len = sel.size(), from = 0, k = 0;

  for (; k < len; ++k) {
    if (k > 0 && (sel[k] == '.' || sel[k] == '#' || sel[k] == ':' || sel[k] == '[')) {
      tab.push_back(sel.substr(from, k-from));
      from = k;
    }
  }

  // last item until end of line
  if (k > from) tab.push_back(sel.substr(from, k-from));

  // NB: il faudrait gerer les repetitions + gestion bizare xxx:checked:active

  for (auto& it : tab) {
    switch (it[0]) {
      case '.':
        spec += GStyle::ClassSpecif;
        name += it;
        break;
      case '#':
        spec += GStyle::IdSpecif;
        name += it;
        break;
      case ':':
      case '[':
        spec += GStyle::ClassSpecif;
        // check specif valid here !!!
        pseudos.push_back(it);
        break;
      default:
        spec += GStyle::TypeSpecif;
        name += it;
        break;
    }
  }

  if (!pseudos.empty()) {
    ps = GPseudoStyle::create(pseudos);
    if (!ps) {
      error("wrong pseudo-class in selector: " + sel);
      return false;
    }
  }
  return !tab.empty();
}


// parses a style, ie: .psclass #id {prop1: val1; prop2:val2; ....}
bool GStyleBase::parseStyle(std::istream& in) {
  GString sel, decl;
  GStrings seltab;

  skipSpaces(in);
  if (in.eof()) return true;
  if (in.fail()) return false;

  int n = getSelector(in, sel, seltab);
  if (n == 0) {
    error("syntax error in selector or missing }: " + sel);
    return false;
  }

  if (!getDeclBlock(in, decl)) {
    error("syntax error in declaration block or missing }: " + sel + " {" + decl + "}");
    return false;
  }

  //cerr << "\nsel: '" << sel << "'"<< endl;
  //cerr << "decl: '" << decl << "'"<< endl;
  return addStyle(seltab, decl);
}


bool GStyleBase::addStyle(GStrings& seltab, GString const& decl) {
  int n = (int)seltab.size();

  if (n == 0 || decl.empty()) return true;  // empty decl OK

  GString name;
  GProp::Specif specif{};
  GPseudoStyle* ps{};

  if (!parseSelName(seltab[0], name, specif, ps)) return false;
  //cerr << "sel: " <<seltab[0] << " ps: " << ps << endl;

  if (n == 1) {
    obtainStyle(name).addStyleDecl(decl, specif, ps);
    return true;
  }

  int desc = 0;
  GStyle* asc = &obtainStyle(name);

  while (true) {

    desc++;
    char reltype = 0;
    bool is_subrule = false;

    if (isValid(seltab[desc])) {
      if (n > desc+1) is_subrule = true;  // still another name after seltab[desc]
    }
    else if (seltab[desc][0] == '>') {   // direct child
      if (n == desc+1) {
        error("missing name in selector before {"+decl+"}");
        return false;
      }
      else {
        reltype = '>';
        desc++;
        if (n > desc+1) is_subrule = true;
      }
    }
    else {
      error("error or not implemented ("+seltab[desc]+") before {"+decl+"}");
    }

    GString descname;
    GProp::Specif descspecif{};
    GPseudoStyle* descps{};

    if (!parseSelName(seltab[desc], descname, descspecif, descps))
      return false;

    specif += descspecif;

    asc = asc->addStyleDecl2((is_subrule ? "" : decl), specif, descps,
                             descname, reltype, is_subrule);

    if (!is_subrule) return true;
  }
}


// note: the name is not parsed: can't be a true selector
void GStyleBase::addCoreStyle(GString const& name, GStyle& s) {
  s.id_ = GCore::obtainId(name);
  s.realized_ = true;
  s.specif_ = GStyle::TypeSpecif;
  //styles_.insert_or_assign(s.id_, &s);
  styles_[s.id_] = &s;
}

}
