//
// Gadget Type
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_GadgetType_hpp
#define Guit_GadgetType_hpp
#include <gtypes.hpp>
namespace guit {

/// Base class of Gadget Types (kind of metaclass).
class GadgetType {
public:
  /// Return the name id, style of the type (depends of subclass).
  /// @{
  GString const& name() const {return name_;}
  GId id() const {return id_;}
  GStyle& style() const {return *style_;}
  /// @}

  /// Instantiates a gadget of this type (redefined in subclasses).
  virtual Gadget& newInstance(GString const& arg = "") = 0;

  /// Instantiates a gadget according to the specified type name.
  /// @warning: returns null if the desired class has not yet been loaded, which can be done
  /// by instantiating a gadget of this class or by calling the gadget's Class() method.
  static Gadget* createInstance(GString const& type_name, GString const& arg);

  /// Returns the type corresponding to this name if it exists, null otherwise.
  static GadgetType* findType(GString const& type_name);

  bool operator==(GadgetType const& other) const;
  bool operator!=(GadgetType const& other) const;

  /// The delete operator does nothing and should not be used.
  void operator delete(void*);

  virtual ~GadgetType() = default;

protected:
  friend class Gadget;
  GadgetType(GString const& name);
  GadgetType(GadgetType&) = delete;
  GadgetType(GadgetType&&) = delete;
  GadgetType& operator=(GadgetType const&) = delete;

  GString name_;
  GStyle* style_{};
  GId id_{};
};


/// GadgetType of a Gadget of class T.
template <class T, class Super>
class GadgetType_ : public Super::GType {
public:
  /// Type of a gadget instantiated by this Type.
  using InstanceType = T;

  T& newInstance(GString const& arg = "") override {return *new T(arg);}

  /// @internal.
  GadgetType_(GString const& name) : Super::GType(name) {}
};


/// @internal.
struct GadgetModes {
  using Type = uint32_t;
  Type dir:2, placement:3, menuLayout:1, unclippedLayers:1, stealTitle:1,
multiLine:1, richText:1,
shown:1, active:1, checked:1, enabled:1, focus:1,
activable:1, downActivable:1, checkable:1, focusable:1, internal:1,
catchEvents:1, ignoreEvents:1, choiceBehavior:2,
hasTip:1, hasCheck:1, hasChoice:1;
};


/// @internal.
struct GadgetSettings {
  using Type = uint8_t;
  Type trulyEnabled:1, hotkeySet:1, modesSet:1, itemsSet:1, hasItems:1,
hasOpener:1, hasMenu:1, hasViews:1;
};


/// @internal.
struct GadgetLayout {
  using Type = uint8_t;
  Type flexType:2, passivePos:1, activePos:1, activeSize:1, reshapeFrame:1,
hasLayers:1;
  uint8_t layer{};
  int8_t flex1{}, flex2{}; //uint8_t wpercent{}, hpercent{};
};


/// @internal.
struct GadgetShape {
  class GBorderValue* border{};
  float start{}, extent{};
};

}
#endif
