//
// GBox: main container gadget
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Box_hpp
#define Guit_Box_hpp
#include <vector>
#include <gadget.hpp>
namespace guit {

/// Container gadget.
/// GBox inherits from Gadget, see Gadget doc for info about properties, callbacks, auto deletion, etc.
/// Contrary to a Gadget, a GBox can contain any other gadget:
/// @code
/// 1.   auto& mybox = Box(".boxes #mybox")
/// 2.   << "My Box"                             // creates and add a GLabel
/// 3.   << Label("is beautiful")                // adds another GLabel
/// 4.   << (Button("Do It") << [&]{doIt();})    // adds a GButton
/// 5.   << (HBox()                              // adds a horizontal box that
/// 6.       << "value: " << TextField()   // contains a GLabel and a GTextField
/// 7.       );
/// @endcode
/// - Box() returns a new GBox instantiated in the heap:  GBox is the C++ class, Box() a helper function
///  that returns *new GBox() (same principle for all gadgets)
/// - The argument of Box() (or of the GBox constr) is a string optionally containing (in this oerder):
///    # one or several CSS classnames ( .boxes )
///    # one CSS id ( #mybox )
///    # an inline CSS (missing here)
///    # the title of the box (missing here, can also be specified in CSS)
/// - At line 2, a string is added to the GBox. It is automatically converted into a gadget which type
///  depends on its parent. As its parent is a GBox it is converted into a GLabel.
/// - A GLabel is also added at line 3 but it is explicitely created: Label() instantiates GLabel
/// - Similarly, at GButton is added at line 4. This GButton contains a callback function that is called
///  on a default condition that depends on its parent (for a GButton: when it is clicked)
/// - Finally a GHBox (horizontal box) is added to the GBox.
///
/// GBox is vertical by default: it stacks it children on top of each other and makes them having
/// the same width. This can be changed by adding a GFlow prop to the box:
///  @code
///    mybox << GFlow::row;   // makes mybox horizontal (as GHBox)
/// @endcode
/// Possible values are: GFlow::row (horizontal), GFlow::column (vertical), GFlow::grid (grid layout),
/// GFlow::pie (pie layout), or GFlow::stack (children are superimposed).
///
/// The children are aligned or justified depending on the GAlign prop of the box and on the
/// GSize and GFlex props of each child.
///
/// Callback functions can be added to the GBox in the same way as with the GButton at line 4.
/// If no on trigger is specified, a default condition (equivalent to the on::action trigger) is used.
/// This condition depends on the gadget: for a GButton it's when it is clicked, for a GBox it's when
/// the default condition of one of its child occurs.
///
class GBox : public Gadget {
protected:
  GBox(GadgetType& c) : Gadget(c) {}
  GBox(GadgetType&, GString const& css_and_title);

public:
  using IsBox = bool;

  virtual ~GBox();
  GBox(GString const& css_and_title = "");
  GBox& clone(GClone const&) override;

  GBox* toBox() override {return this;}
  GBox const* toBox() const override {return this;}

  /// Returns the prop that controls the exclusive selection of children, if any.
  /// returns null if this box does not enforce exclusive selection.
  /// @{
  virtual GChoice* choiceProp();
  virtual GProp* valueProp() override;
  /// @}

  enum Remove {
    InContent = 1<<0, InBorder = 1<<1, Internals = 1<<2
  };

  /// Searches if the gadget contains a child gadget of (exactly) this class.
  /// The first version returns the gadget as a Gadget* if it is found, and null otherwise.
  /// In addition, the second version casts the returned gadget to the desired type.
  /// @note This method does not take into account derived types.
  /// @{
  virtual Gadget* findChild(GadgetType& type, bool in_border = false) const;

  template <class T>
  T* findChild(bool in_border = false) const {
    return dynamic_cast<T*>(findChild(T::Type(), in_border));
  }
  /// @}

  /// Returns the child at this position or at this place.
  /// - if pos is one of Gadget::Place, returns the child at this place in the border
  /// - otherwise, returns the child at this position in in content area (0 is first child, -1 last child).
  /// @return childAt() returns the corresponding child, or null
  /// @return operator[] returns the corresponding child, or throws a GFatalError exception,
  /// which resets the event loop.
  /// @{
  virtual Gadget* childAt(int pos) const;
  Gadget& operator[](int pos) const;
  /// @}

  /// Returns the position or the place of this child.
  /// Returns -1 if  the child does not belong to this box.
  /// @{
  virtual int childPos(Gadget* child) const;
  int childPos(Gadget& child) const {return childPos(&child);}
  /// @}

  /// Returns the direct (or indirect) child containing this point.
  /// - if deep is false (the default), returns the direct child containing this point
  /// - if deep is true, returns the smallest direct or indirect child containing this point
  /// - If _pred_ is not null, children that dont verify this predicate are ignored
  /// The point is relative to *this* when using childAtPoint() and to its window when using childAtWPoint()
  /// @{
  virtual Gadget* childAtPoint(GPoint const& point, bool deep = false,
                               GPredicate const& pred = {});
  virtual Gadget* childAtWPoint(GPoint const& wpoint, bool deep = false,
                                GPredicate const& pred = {});
  /// @}

  /// Returns the postion of the closest child from this position (in window coordinates).
  virtual int childAroundWPoint(GPoint const& wpoint, bool localpos);

  /// Returns the content aread child list.
  /// @note Gadgets loacted in the borders (e.g. menubars, scrollbars) are stored in the border list.
  /// @note Windows and Menus are stored in the prop list
  /// @{
  GadgetList& children() {return children_;}
  const GadgetList& children() const {return children_;}
  /// @}

  /// Returns the number of children if the content area.
  int childCount() const;

  using Gadget::contain;

  /// Returns true if the first argument is a direct (or indirect) child of the box.
  /// If indirect is  false (the default) searches only for direct children, otherwise searches for
  /// direct and indirect children.
  /// @{
  virtual bool contain(Gadget* child, bool indirect = false) const;
  virtual bool contain(Gadget& child, bool indirect = false) const {
    return GBox::contain(&child, indirect);
  }
  /// @}

  /// Calls a function on all children in the content area.
  /// _fun_() is called on each element in the childlist.
  /// @{
  virtual void perform(std::function<void(Gadget& child)> fun);
  virtual void perform(std::function<void(Gadget& child, int pos)> fun);
  /// @}

  using Gadget::add;

  /// Adds a gadget to the box.
  /// operator<<() provides a more convenient API:  a << b << c  adds b and c to a.
  /// The appropriate adding function is automatically called.
  ///
  /// However, contrary to operator<<(), this method allows specifying the position in the box' content area
  /// or the place in its border:
  /// - if pos is one of Gadget::Place, this method replaces the child at this place in the box' border
  /// - otherwise, it inserts the child at the specifield position in in content area (0 is first child, -1 last child).
  ///
  /// If _child_ is a non-window, non-menu gadget that belongs to another box, it is removed from it.
  /// If it already belongs to this box it is moved to the new position.
  ///
  /// If _child_ is a window/menu/dialog box  (GWindow, GDialog, GMenu), it is not removed from other
  /// parents because such widgets can be added to multiple parents (including non-box gadgets).
  /// @{
  virtual void add(Gadget* child, int pos = -1);
  virtual void add(Gadget& child, int pos = -1) {GBox::add(&child, pos);}
  /// @}


  /// Adds a Gadget child that is auto-created.
  /// The string is converted into a gadget that is added to the childlist (see add(Gadget*, int index)).
  /// The type of the created gadget depend on the type of the box, for instance:
  /// # Ordinary boxes (GHBox, GHPane, etc.) and tables (GTable, GTRow, GTHead) create GLabel gadgets.
  /// # GChoiceBox boxes and their subclasses create GItem gadgets
  /// # GChart boxes create GChartItem gadgets
  /// @{
  void add(GString const& item) override;
  void add(GString const& item, int index);
  void add(GStrings const& itemlist);
  /// @}

  using Gadget::remove;
  
  /// Removes a child.
  /// Returns true if the child was actually removed (i.e. if it belonhs to *this*), and false otherwise.
  /// @note Removed children are deleted if they are not gptr-pointed elsewhere (see gptr<> smart pointers)
  /// @{
  virtual bool remove(Gadget* child);
  virtual bool remove(Gadget& child) {return GBox::remove(&child);}
  /// @}

  /// Removes a child at a given position or a given place.
  /// - if pos is one of Gadget::Place, removes the child at this place in the border
  /// - otherwise, removes the child at this position in in content area (0 is first child, -1 last child).
  virtual bool removeAt(int pos);

  /// Removes all children.
  /// Remove all children iin the content area and/or in the border depending on argument.
  virtual void removeChildren(Remove mode = Remove::InContent);
  
  /// Removes all Props and children.
  /// Calls removeProps() then removeChidren()
  void cleanup() override;

  enum struct ChoiceBehavior : uint8_t {
    NoChoice=0, Select=1, Show=2
  };

  /// Returns the GChoiceBehavior of the box.
  ChoiceBehavior choiceBehavior() const {return ChoiceBehavior(modes_.choiceBehavior);}

  void setValue(GString const&) override;
  void clearValues() override;

  void doPaint(GRenderGraphics&, GRender* parent_render) override;
  Gadget* doPick(GPoint const&, class GPickMode&) override;

  using GType = GadgetType_<GBox,Gadget>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// Guit provides full access to all fields, but beware using them is at your own risk.
  /// @{
  virtual void topInit();
  virtual void initChildren(GRender* parrender);
  void initClone(Gadget& from, GClone const&) override;
  void addProp(GProp&, bool add_to_proplist) override;
  virtual void addItem(class GItemSpec const& item);
  Gadget* addItemImpl(GString const& item, int index, GString& defclass);
  virtual Gadget& createDefaultItem(GString const& arg);
  virtual void setChoiceImpl(GChoice& c, bool add_to_list);

  virtual GDim computeInsideSize(GRender&);
  virtual void computeInsideSizeColumn(GDim& boxsize, GRender&);
  virtual void computeInsideSizeRow(GDim& boxsize, GRender&);
  virtual void computeInsideSizeStack(GDim& boxsize, GRender&);
  GRect childrenArea(GRender&) const;

  void pack(GRender* parent_render) override;
  virtual void packColumn(GRect& ch_area, GRender&);
  virtual void packRow(GRect& ch_area, GRender&);
  virtual void packPie(GRect& ch_area, GRender&);
  virtual void packStack(GRect& ch_area, GRender&);
  virtual void packBorders(GRect& ch_area, GRender&); // changes ch_area
  void moveImpl(float winx, float winy, bool callcb) override;

  struct BorderChild {
    BorderChild(Gadget* c, Place p) : child_(c), place_(p) {}
    gptr<Gadget> child_;
    Place place_;
  };
  using BorderList = std::vector<BorderChild>;
  GadgetList children_;
  BorderList* bchildren_{};
  void borderAdd(Gadget& child, Place place);
  BorderList* borderChildren() {return bchildren_;}
  BorderList const* borderChildren() const {return bchildren_;}
  /// @}
};


/// Vertical Box (alias of GBox).
using GVBox = GBox;

/// Horizontal Box (variant of GBox).
class GHBox : public GBox {
public:
  GHBox(GString const& css_and_title = "") : GBox(Type(), css_and_title) {}

  using GType = GadgetType_<GHBox,GBox>;
  static GType& Type();
  GType& type() const override {return Type();}
};


/// Box with a rounded border; see GBox.
class GPane : public GBox {
public:
  GPane(GString const& css_and_title = "") : GBox(Type(), css_and_title) {}

  using GType = GadgetType_<GPane,GBox>;
  static GType& Type();
  GType& type() const override {return Type();}
};



/// helper functions that instantiate a GBox, a GVBox, a GHBox, a GPane in the heap.
/// @{
GBox&  Box(GString const& css_and_title = "");
GVBox& VBox(GString const& css_and_title = "");
GHBox& HBox(GString const& css_and_title = "");
GPane& Pane(GString const& css_and_title = "");
/// @}



/// Box containing buttons (GBox variant).
/// Strings added to this gadget are converted into push buttons.
/// Default orientation is vertical, this can be changed by adding a GFlow prop.
/// @code
///    auto& box = ButtonBox
///    << Label("A Label")                // creates a Label
///    << ("Do It" << [this]{doIt();})    // creates a Button
/// @endcode
/// @see GBox and Gadget.
class GButtonBox : public GBox {
protected:
  GButtonBox(GadgetType&, GString const& css_and_title);

public:
  GButtonBox(GString const& css_and_title = "");

  Gadget& createDefaultItem(GString const& arg) override;

  using GType = GadgetType_<GButtonBox,GBox>;
  static GType& Type();
  GType& type() const override {return Type();}
};

GButtonBox& ButtonBox(GString const& css_and_title = "");


}
#endif
