//
//  Sliders & Scrollbars
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gevent.hpp>
#include <gslider.hpp>
#include <gscrollbox.hpp>
#include <gtext.hpp>
#include <gnumber.hpp>
#include <gcontrol.hpp>
#include <gstyle.hpp>
#include <grender.hpp>
using namespace std;
namespace guit {

static const float ScrollBarMinSize = 20.f;

GIMPLEMENT_GADGET_ALL(Slider, GValueItem)
GIMPLEMENT_GADGET_TYPE(ScrollBar)

GSlider::GSlider(GadgetType& t, GString const& spec) : GValueItem(t, spec) {}

GSlider::GSlider(float value) : GValueItem(Type(),"") {
  setValue(value);
}

GSlider& Slider(float value) {return *new GSlider(value);}

GSlider& GSlider::setRange(float min, float max) {
  value_->setRangeF(min, max);
  return *this;
}

GFlow::Kind GSlider::getFlow() {
  return GFlow::Kind(modes_.dir);
}

void GSlider::paintContent(GRenderGraphics& gr, GRender& r, GFlow::Kind dir) {
  auto area = frame_;
  float knobsize{};

  switch (dir) {
    case GFlow::Row:
      knobsize = area.h;
      area.x += (frame_.w - knobsize) / frame_.w * offset_;
      area.w = knobsize;
      area.h = knobsize;
      // cerr << "paint " << this
      // << " / " << area.x << " " << frame_.w << " " << offset_ << endl;
      break;
    case GFlow::Column:
      knobsize = area.w;
      area.y += (frame_.h - knobsize) - (frame_.h - knobsize) / frame_.h * offset_;
      area.w = knobsize;
      area.h = knobsize;
      break;
    default:
      break;
  }
  gr.drawOval(area);
  if (isActive()) {
    gr.unsetBgColor();
    gr.setColor(GColor::guitblue);
    gr.drawOval(area);
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GScrollBar::GScrollBar(GadgetType& t, GString const& spec) : GSlider(t, spec) {
  setRange(0.f, 1.f);
}

GScrollBar::GScrollBar(GString const& spec) : GScrollBar(Type(), spec) {
  setRange(0.f, 1.f);
}

GScrollBar& GScrollBar::clone(GClone const& c) {
  auto* g = new GScrollBar;
  //if (c & GClone::ShareProps) g->value_ = value_;  // done by initClone()
  g->initClone(*this, c);
  return *g;
}

void GScrollBar::paintContent(GRenderGraphics& gr, GRender& r, GFlow::Kind dir) {
  auto area = frame_;
  GScrollBox* sbox{};
  Gadget* child{};
  if (auto* par = getParent()) {
    if ((sbox = par->toScrollBox())) child = sbox->childAt(0);
  }
  if (!child) return;

  switch (dir) {
    case GFlow::Row: {
      auto sz = frame_.w * frame_.w / child->w();
      if (sz > frame_.w) sz = frame_.w;
      area.x += (frame_.w - sz) / frame_.w * offset_;
      area.w = sz > ScrollBarMinSize ? sz : ScrollBarMinSize;
      //area.h = knobsize;
    } break;

    case GFlow::Column: {
      auto sz = frame_.h * frame_.h / child->h();
      if (sz > frame_.h) sz = frame_.h;
      area.y += (frame_.h - sz) / frame_.h * offset_;
      //area.w = knobsize;
      area.h = sz > ScrollBarMinSize ? sz : ScrollBarMinSize;
    } break;

    default:
      break;
  }

  gr.drawRoundRect(area, 4, 4);
  if (isActive()) {
    gr.unsetBgColor();
    gr.setColor(GColor::guitblue);
    gr.drawRoundRect(area, 4, 4);
  }
}


void GScrollBar::mouseDrag(GMouseEvent& e) {
  if (!isTrulyEnabled()) return;  // || !isEdited

  switch (getFlow()) {
    case GFlow::Row:
      value_->increaseF(e.dx() * (max() - min()) / frame_.w);
      return;
    case GFlow::Column:                 // inverted !!
      value_->increaseF(e.dy() * (max() - min()) / frame_.h);
      return;
    case GFlow::Pie:
      // TODO
      return;
    default:
      return;
  }
}

}
