//
//  Base class for Windows and Dialogs.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Window_hpp
#define Guit_Window_hpp
#include <gbox.hpp>
namespace guit {

/// Window Gadget.
/// See GDialog and subclasses for a window with OK, Cancel buttons.
/// GDialog also provides static methods for opening predefined dialogs.
///
/// The title and the icon can be specified by adding a GTitle and a GIcon prop (or a 'title' and 'icon'
/// prop in CSS files)
///
/// ### Key methods:
/// - startApp() : starts the application event loop. This method is blocking and should be called at last.
///  The window then becomes the *main window* of the application
/// - setModal(state) : makes the window modal or not (not modal by default)
/// - show(state) : shows/hides the window.
/// - close() : called when the close button of the window bar, or the OK, Cancel buttons of a GDialog
///  are clicked. This method may delete the window as explained below.
///
/// ###  Closing the window
/// close() performs the following operations:
/// - If a GQuitDialog and/or on::close triggers were added to the window:
///   * the close triggers are fired (the corresponding callbacks are executed)
///   * the QuitDialog is opened (example below).
/// Otherwise:
///   * if the window is the *main window* (see startApp() above), the application terminates
///   * otherwise the window is hidden (by calling show(false)) and deleted if it is in dieOnClose
///    mode (see dieOnClose()), has no parent and is not gptr-pointed.
///
/// Example: quitdialog will be shown when the window is closed:
/// @code
///    auto& quitdialog = QuitDialog("Changes will be lost");
///    auto& win = Window("My Window")
///    << quitdialog
///    << etc...;
/// @endcode
///
/// ### Triggers:
/// - on::show : when the window is shown (e.g. when show(true) is called)
/// - on::hide : when the window is hidden (e.g. when show(false) is called)
/// - on::done : when the window is being closed (i.e. when close() is called)
/// - on::action (GDialog subclass) : when the OK button is clicked.
///
class GWindow : public GBox {
protected:
  GWindow(GadgetType&, GString const& css_and_title);

public:
  GWindow(GString const& css_and_title = "");
  ~GWindow();

  GWindow& clone(GClone const&) override;
  GWindow* toWindow() override {return this;}

  /// Starts the event loop
  /// Same effect as GApp::start() with this window being the main window.
  /// This function is blocking and must be called last.
  virtual int startApp();

  /// Returns true if the window is modal.
  virtual bool isModal() const {return wmodes_.modal;}
  
  /// Makes this window modal (arg is true) or not modal (arg is false).
  virtual void setModal(bool modal);

  /// Shows/hides the window.
  /// Note that show(true) raises the window (see raise()).
  GWindow& show(bool show = true) override;

  virtual void show(GEvent&, bool autoplace);

  /// Closes the window.
  /// Called when the close button of the window bar is clicked
  /// or the OK, Cancel buttons of a GDialog subclass.
  ///
  /// Fires the on::done trigger callbacks and cond props, then:
  /// - does nothing if dieOnClose(true) was previously called
  /// - otherwise, terminates the app if this is the main window,
  /// - otherwise, hides the window and detetes if not gptr-pointed elsewhere.
  virtual void close();
  
  /// Specifies whether the window should be deleted when close() is called.
  /// @note even if this function was called, the window will not be deleted if it is gptr-pointed
  /// elsewhere (ie pointer by a gptr<> pointer or by a parent gadget).
  virtual void dieOnClose(bool dies = true);

  /// find the window containg this point.
  static GWindow* findWindow(GPoint const& screenpos);

  GPoint pos() const override {return screenpos_;}
  float x() const override {return screenpos_.x;}
  float y() const override {return screenpos_.y;}

  virtual void raise();
  virtual void minimize();
  virtual void maximize();
  virtual void restore();

  GWindow& setResizable(bool);
  GWindow& setMinSize(float w, float h);
  GWindow& setMaxSize(float w, float h);
  GWindow& setOpacity(float opacity);  ///< opacity is between 0.f and 1.f.
  GWindow& stayOnTop();
  GWindow& stayOnTopOf(GWindow&);
  GWindow& joinAllSpaces();
  GWindow& setBordered(bool);

  void setScale(float scale);
  float scale() const {return scale_;}

  using GBox::add;
  void add(Gadget* child, int index = -1) override;
  void add(Gadget& child, int index = -1) override {GWindow::add(&child, index);}
  void add(GWindow*) override;

  bool isTopLevel() const override {return true;}
  void update() override;
  void updateLayout(Update = Update::IfNeeded) override;

  class GNatWindow* nativeWindow() {return natwin_;}

  using GType = GadgetType_<GWindow,GBox>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// @{
  uint32_t windowID() const;
  void topInit() override;
  void addProp(GProp& prop, bool add_to_list) override;
  void moveImpl(float winx, float winy, bool callcb) override;
  void addHotkey(Gadget&, class GHotkey&);
  bool triggerHotkey(GEventState keystate, GKeyCode key);
  void setIconImpl(GString const& filename);

  struct WinModes {uint8_t
  modal:1, mustupdate:1, mustdraw:1, mustlayout:1, dieonclose:1,
  autoclose:1, softwin:1, quitcalled:1;
  } wmodes_{};
  float scale_{1.f};
  GPoint screenpos_;
  class GNatWindow* natwin_{};
  class GHotkeys* hotkeys_{};
  /// @}
};

/// Creates a GWindow.
GWindow& Window(GString const& css_and_title = "");

}
#endif
