//
//  Guit Sockets and Socket Channels.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Socket_hpp
#define Guit_Socket_hpp
#include <map>
#include <gtypes.hpp>
#include <gproplist.hpp>
namespace guit {

namespace impl {
class GSocketBase;
class GSocketCnx;
class GSocketServer;
class GSocketNotifier;
}
class GSocket;


/// A Channel of communication that can be bound with a GVarProp both ways.
/// channel["*"] is a special channel that sends/receives all data unformatted.
class GSocketChannel : public GObject {
public:
  /// Output binding: when _sender_ value will change, this value will be sent on this channel.
  GSocketChannel& operator<<=(GVarProp& sender);
  
  /// Input binding: when data will arrive on this channel, _receiver_ value will be updated.
  /// @note when data is sent
  /// @{
  GSocketChannel& operator>>=(GVarProp& receiver);
  friend GSocketChannel& operator<<=(GVarProp& receiver, GSocketChannel&);
  /// @}

  /// Binds both ways.
  GSocketChannel& operator^=(GVarProp& synced_prop);

  /// Adds a callback function that is fired when the channel receives data.
  GSocketChannel& operator<<(GFunction const& fun);

  /// Sends data on this channel.
  GSocketChannel& operator=(GString const& data);
  
  /// Returns the socket that is bound to this channel.
  GSocket* socket() {return sock_;}

  /// Returns callbacks.
  const GPropList& props() const {return props_;}

  GString typeName() const override {return "SocketChannel";}

private:
  friend class GSocket;
  friend class impl::GSocketCnx;
  friend class impl::GSocketNotifier;
  GString name_;
  gptr<GVarProp> recv_, sender_;
  gptr<GSocket> sock_;
  GPropList props_;
  void sendValue();
};



/// Guit Socket; can either be a Server or a Client depending on it is started.
/// After creating the GSocket, startAsServer() starts it as a server and startAsClient() as a client.
/// A GSocket can manage an abritary set of GSocketChannel(s). Each of them is treated as
/// a separate communication channel, and can be bound with a mutable Property (GVarProp).
/// @note GSocket rely on IPV4.
class GSocket : public GObject {
public:
  /// Server Mode.
  enum Mode {Unstarted, Client, Server};
  
  /// Socket Status.
  /// - Invalid: invalid socket or wrong socket type
  /// - CannotReach: can't connect (Client), can't bind (Server)
  /// - UnknownHost: can't find host
  enum Status {
    Invalid = -3, UnknownHost = -2, CannotReach = -1,
    Disconnected = 0, Connected = 1
  };

  GSocket();
  ~GSocket();

  /// Returns socket mode (ie server or client, see enum Mode).
  Mode mode() const {return mode_;}

  /// Returns connexion status (see enum Status).
  GInt& status();

  /// Starts the GSocket as a Server socket.
  /// Returns true if the server socket could be started (check status() in case of an error).
  bool startAsServer(int port);
  
  /// Starts the GSocket as a Client socket.
  /// Returns true if the client socket could be started (check status() in case of an error).
  bool startAsClient(GString const& hostname, int port);
  
  /// Disconnects the socket.
  void disconnect();

  /// Returns a channel of communication that can be bound to a mutable Property.
  /// mysocket["chan"] returns the channel named "chan". Data sent to this channel will be
  /// received by the channel with the same name on the other side, and vice-versa.
  /// The name must be an ASCII string without spaces nor special characters.
  ///
  /// mysocket["*"] is a special channel that sends/receives all data unformatted. Other channels
  /// prefix data by "_name_:"
  ///
  /// Callbacks can be added to each channel, and  channels can be bound with mutable
  /// properties; see GSocketChannel for details.
  GSocketChannel& operator[](GString const& name);

  /// Specifies whether data is sent back while receiving data.
  /// Useful for avoiding useless data exchanges, especially when props are synced on both sides.
  /// When props are synced (or if both sides have callbacks) changing a updates b that updates a.
  /// If a has not been changed in the meantime everything is fine, otherwise this may cause jitter
  /// or infinite loops.
  void setSendBack(bool state) {dontsendback_ = !state;}

  GString typeName() const override {return "Socket";}

protected:
  friend class GSocketChannel;
  friend class impl::GSocketServer;
  friend class impl::GSocketCnx;
  
  void reset();
  void clear();
  void update();
  void sendData(GString const&);
  void disconnected(impl::GSocketBase*, GString const&);
  
  /// the socket is disconnected if it receives nothing more than this amount of times.
  int maxattempts_{50};
  Mode mode_{};
  bool dontsendback_{};
  gptr<GInt> status_;
  std::map<GString, gptr<GSocketChannel>> channels_;
  impl::GSocketBase* impl_{};
};

}
#endif
