//
//  Triggers
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Trigger_hpp
#define Guit_Trigger_hpp
#include <forward_list>
#include <gcond.hpp>
namespace guit {

/// Specifies when conditional statements are executed.
/// @see GCond.
class GTrigger {
public:
  static GTrigger
  /// no action.
  none,
  /// default action (meaning depends on gadget).
  action,
  /// the action was done (meaning depends on gadget).
  done,
  /// the value has changed.
  change,
  /// the value is going to be changed.
  changing,
  /// the gadget was checked (selected).
  checked,
  /// the gadget was unchecked (deselected).
  unchecked,
  /// click on a gadget.
  click,
  /// double click on a gadget.
  dblclick,
  /// mouse press on a gadget.
  /// if the gadget does not accept button events (e.g. Labels) the mousedown is sent to its parent.
  mousedown,
  /// mouse release on a gadget.
  /// the gadget that receives mouseup is the gadget that received mousedown
  mouseup,
  /// mouse moved with at least one button pressed.
  /// the gadget that receives mousedrag is the gadget that received mousedown
  mousedrag,
  /// mouse moved with no button pressed.
  /// the gadget that receives mousemove is the gadget beneath the mouse cursor.
  /// @note callbacks can also be added to mouse cursors (see GCursor).
  mousemove,
  /// the mouse enters a gadget.
  /// the gadget that receives mouseenter is the gadget that was entered.
  mouseover,
  /// the mouse leaves a gadget.
  /// the gadget that receives mouseleave is the gadget that was exited.
  mouseout,
  /// the mouse wheel rolls up or down.
  /// the gadget that receives wheel is the closest gadget beneath the mouse cursor
  /// that accepts wheel events.
  wheel,
  /// a key was pressed.
  /// the gadget that receives keydown is the gadget that has the focus.
  keydown,
  /// a key was released.
  /// the gadget that receives keydup is the gadget that has the focus.
  keyup,
  /// text was entered.
  /// the gadget that receives keytype is the gadget that has the focus.
  keytype,
  /// the text caret was moved.
  /// the gadget that receives caret is the gadget that has the focus.
  caret,
  /// a touch gesture was performed.
  /// the main window receives the event.
  touch,
  /// a multitouch gesture was performed.
  /// the main window receives the event.
  gesture,
  /// the gadget got the keyboard focus.
  focus,
  /// the gadget left the keyboard focus.
  focusout,
  /// the gadget was enabled.
  enabled,
  /// the gadget was disabled.
  disabled,
  /// the gadget was shown.
  show,
  /// the gadget was hidden.
  hide,
  /// the gadget was moved.
  move,
  /// the gadget was initialized (eg the CSS was set).
  init,
  /// the gadget was resized.
  resize,
  /// the gadget was repaint.
  paint,
  /// an element (prop, callback, gadget) was added to this gadget.
  add,
  /// an element (prop, callback, gadget) was removed from this gadget.
  remove,
  /// text was copied.
  copy,
  /// text was cut.
  cut,
  /// text was pasted.
  paste,
  /// custom user event.
  user,
  /// impl.
  changeaction;
  
  using Filter = int32_t;

  virtual ~GTrigger() = default;

  /// Constr.
  /// _mask_ is an ORed mask of the event types need to be activated.
  GTrigger(GEventType mask, Filter filter) : mask_(mask), filter_(filter) {}
 
  /// Constr.
  GTrigger(GTrigger const& from) : mask_(from.mask_), filter_(from.filter_) {}

  /// Returns the event mask.
  GEventType eventMask() const {return mask_;}
  
  /// Returns the filter.
  Filter filter() const {return filter_;}

  /// Returns true if the triggers are equivalent.
  virtual bool operator==(GTrigger const& other) const;

  /// Returns a guard (see GGuard).
  class GGuard& operator[](GExprBool const& guard_expr);

  static GTrigger* getTrigger(GString const& trigger_name);

  /// @internal
  /// @{
  virtual bool isEventTrigger() const {return true;}
  virtual class GExprTrigger* toExprTrigger() {return nullptr;}
  virtual const class GExprTrigger* toExprTrigger() const {return nullptr;}

  virtual void receiverAdded(Gadget&, GCond&);
  virtual void receiverRemoved();
  virtual void receiverDeleted() {}

  static GEvent* curevent_;
  GEventType mask_{};
  Filter filter_{};
  
protected:
  GTrigger() = default;
  /// @}
};

/// Shortcut for GTrigger.
using on = GTrigger;


// Trigger that depends on an active expression, a state or a model.
class GExprTrigger : public GTrigger {
public:
  ~GExprTrigger();

  // creates a trigger that is fired when the value of the expr becomes true
  GExprTrigger(class GBoolExpr&);

  bool operator==(GTrigger const& other) const override;

  bool isEventTrigger() const override {return false;}
  GExprTrigger* toExprTrigger() override {return this;}
  const GExprTrigger* toExprTrigger() const override {return this;}

  void fire(GObject& sender, bool active);
  GNotifier* notifier() const {return notifier_;}
  void setNotifier(GNotifier*);
  void addSender(GObject& sender);
  void senderRemoved(GObject&);
  void receiverAdded(Gadget&, GCond&) override;
  void receiverRemoved() override;
  void receiverDeleted() override;
  Gadget* receiver() {return recv_;}

protected:
  friend class GEvent;
  std::forward_list<GObject*> senders_;
  Gadget* recv_{};
  GNotifier* notifier_{};
  GCond* firedprop_{};
  gptr<GBoolExpr> boolexpr_;
  bool deleting_{}, invalid_{};
};

}
#endif
