//
//  Corrects a bug when resizing windows with SDL + more
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
#ifdef __APPLE__

#include <Cocoa/Cocoa.h>
#include <SDL_syswm.h>
#include <gnative.hpp>
using namespace std;

// corrects a bug in SDL
@interface WinDelegate : NSObject<NSWindowDelegate> {
@public
  NSWindow* window_;
  guit::GNatWindow* natwin_;
}
@end

@implementation WinDelegate

- (id)init:(NSWindow*)window natwin:(guit::GNatWindow*)natwin {
  if ((self = [super init])) {
    window_ = window;
    natwin_ = natwin;
    [[NSNotificationCenter defaultCenter]
     addObserver:self
     selector:@selector(windowDidResize:)
     name:NSWindowDidResizeNotification object:window];
  }
  return self;
}


// corrects SDL bug (this function is not called in SDL current implementation)
// permet de rafraichir la fenetre prendant qu'on la retaille
// nb: taille avec titre alors que GNatWindow::resizeCB sans titre
//
- (void)windowDidResize:(NSNotification *)notification {
  /*
  cerr << "windowDidResize "  << &(natwin_->window_)
  << " : " << window_.frame.size.width
  << " " << window_.frame.size.height << endl;
   */
  // [sdldelegate_ windowDidResize:notification];
  natwin_->resizeCB(guit::GNatContext::elapsedTime(),
                    uint32_t(window_.frame.size.width),
                    uint32_t(window_.frame.size.height));
  natwin_->paint();
}

@end


void guit::GNatWindow::createDelegate() {
  SDL_SysWMinfo info;
  SDL_VERSION(&info.version);
  if (!SDL_GetWindowWMInfo(sdlwin_, &info)) {
    gerror("SDL","Could not initialize Cocoa window delegate'");
    delegate_ = nil;
  }
  else {
    delegate_ = [[WinDelegate alloc] init:info.info.cocoa.window natwin:this];
  }
}

void guit::GNatWindow::deleteDelegate() {
  delegate_ = nil;              // should also removeObserver !!!!
}

id guit::GNatWindow::cocoaWindow() {
  return ((WinDelegate*)delegate_)->window_;
}

void guit::GNatWindow::stayOnTop() {
  //NSScreenSaverWindowLevel
  //NSStatusWindowLevel;
  //NSMainMenuWindowLevel
  ((WinDelegate*)delegate_)->window_.level = NSMainMenuWindowLevel+1;
  //auto l = [((WinDelegate*)delegate_)->window_  level];
  //cerr << "stayOnTop level " <<l <<endl;
}

void guit::GNatWindow::stayOnTopOf(GNatWindow& w) {
  auto l = ((WinDelegate*)w.delegate_)->window_.level;
  ((WinDelegate*)delegate_)->window_.level = l+1;
}

void guit::GNatWindow::joinAllSpaces() {
  ((WinDelegate*)delegate_)->window_.collectionBehavior |= NSWindowCollectionBehaviorCanJoinAllSpaces;
}
/*
  // show windows on all virtual desktops
  helpWin.collectionBehavior = NSWindowCollectionBehaviorCanJoinAllSpaces;
  settingsWin.collectionBehavior = NSWindowCollectionBehaviorCanJoinAllSpaces;
  editWin.collectionBehavior = NSWindowCollectionBehaviorCanJoinAllSpaces;
*/

//NSWindowStyleMaskResizable


void guit::GNatWindow::disallowFullScreen() {
  ((WinDelegate*)delegate_)->window_.collectionBehavior
  &= ~(NSWindowCollectionBehaviorFullScreenPrimary | NSWindowCollectionBehaviorFullScreenAuxiliary);
       
  ((WinDelegate*)delegate_)->window_.collectionBehavior
  |= NSWindowCollectionBehaviorFullScreenNone;

  //((WinDelegate*)delegate_)->window_.maxSize = NSMakeSize(300,300);
  //((WinDelegate*)delegate_)->window_.maxFullScreenContentSize = NSMakeSize(10,10);
}

void guit::GNatWindow::setTitlebarTransparent(bool val) {
  ((WinDelegate*)delegate_)->window_.titlebarAppearsTransparent = val;
}

void guit::GNatWindow::hideCloseButton(bool val) {
  [[((WinDelegate*)delegate_)->window_
    standardWindowButton:NSWindowCloseButton] setHidden:val];
}

void guit::GNatWindow::hideMiniaturizeButton(bool val) {
  [[((WinDelegate*)delegate_)->window_
    standardWindowButton:NSWindowMiniaturizeButton] setHidden:val];
}

void guit::GNatWindow::hideZoomButton(bool val) {
  [[((WinDelegate*)delegate_)->window_
    standardWindowButton:NSWindowZoomButton] setHidden:val];
}


void guit::GNatContext::moveCursor(GPoint const& p) {
  CGPoint pos{p.x, p.y};
  CGWarpMouseCursorPosition(pos);
}

void guit::GNatContext::showCursor(bool state) {
  // note that there is a 'hide cursor count': CGDisplayShowCursor shows
  // the mouse cursor if the count is 0.
  //std::cerr << "Cocoa showCursor "<< state << std::endl;
  if (state) CGDisplayShowCursor(kCGNullDirectDisplay);
  else CGDisplayHideCursor(kCGNullDirectDisplay);
}

void guit::GNatContext::freezeCursor(bool state) {
  CGAssociateMouseAndMouseCursorPosition(!state);
}

#endif
