//
// Animations
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <cmath>
#include <ganim.hpp>
#include <gtrigger.hpp>
#include <gevent.hpp>
#include <gtext.hpp>
#include <gitem.hpp>
#include <gwindow.hpp>
#include <gcolor.hpp>
#include <gpos.hpp>
#include <gsize.hpp>
#include <gstyle.hpp>
#include <gcontrol.hpp>
#include <gstring.hpp>
#include <gadgetstyle.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP_TYPE(Anim, GPropType::Anim)

// mis la pour permettre de compiler sans les anims
void GVarProp::blink(GString const& msg, GTime duration) {
  perform([&msg,duration](Gadget&g){g.blink(msg,duration);});
}

void Gadget::blink(GString const& msg, GTime delay) {
  auto* win = getWindow();
  if (!win) return;

  auto* g = this;
  // blink the parent if g not visible (and its pseudo-arent if it's a menu/window)
  if (!g->isShown()) g = g->getPseudoParent();
  if (!g) return;
  g->updateLayout(Update::Now);
  
  auto& anim = Anim(delay, 1)
  << (AnimStep() << Background(1.f,0.f,0.f,0.4f)  << GColor::black)
  << (AnimStep() << Background(1.f,0.f,0.f,0.05f) << Color(0.f,0.f,0.f,0.2f));

  auto& overlay = Label(msg)
  << new GSize(g->w(), g->h(), GSize::Pixels, GSize::Pixels)
  << new GPos(g->frame_.x, g->frame_.y)
  << anim;
  
  // caution: add to the win, not to the central area!
  win->add(overlay);
  anim << on::done /[&overlay,win,&anim]{win->remove(overlay);};
  anim.start();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// NOTE: CSS dont work because addProp isnt called !!!!

GAnim::~GAnim() {
  //cerr << "~GAnim "<< this <<" / timer: "<< timer_ << endl;
  for (auto* step : steps_) delete step;
}

GAnim::GAnim(GTime duration, int count)
: duration_(duration), repeat_(count) {}

GAnim::GAnim(GString const& format) {
  GString dur, rep, rest, props;

  auto n = strsplit(format, dur, rest);
  if (n == 0) return;
  duration_ = guit::toInt(dur);

  n = strsplit(rest, rep, props);
  if (n == 0) return;
  repeat_ = guit::toInt(rep);
  
  std::vector<GString> v;
  strsplit(props, v, ",");
  for (auto& it : v) *this << it;

  start(); // auto-started!
}

GString GAnim::stringValue() const {return "<Anim>";}

GAnim& GAnim::setRepetitions(int count) {
  repeat_ = count;
  omodes_.initialized = 0;
  return *this;
}

bool GAnim::isRunning() const {
  return !timer_ ? false : timer_->isRunning();
}

GAnim& GAnim::operator<<(GFunction const& fun) {
  calls_.addEnd(on::changeaction / fun);
  return *this;
}

GAnim& GAnim::operator<<(GAnimStep& as) {
  steps_.push_back(&as);
  return *this;
}

GAnim& GAnim::operator<<(GProp* prop) {
  if (!prop) {error("operator<<","Null argument"); return *this;}
  else return *this << *prop;
}

GAnim& GAnim::operator<<(GProp& prop) {
  if (prop.toCond()) calls_.addEnd(prop);
  else {
    auto& as = *new GAnimStep() << prop;
    steps_.push_back(&as);
  }
  return *this;
}

GAnim& GAnim::operator<<(GString const& spec) {
  if (spec.empty()) return *this;
  GPropList pl;

  if (spec[0] != '{') {
    GadgetStyle::parseProps(nullptr, &pl, spec, ";");      // A REVOIR !!!!
  }
  else {
    GString inline_style, classes_and_id, title;
    GadgetStyle::parseStyle(spec, classes_and_id, inline_style, title);
    GadgetStyle::parseProps(nullptr, &pl, inline_style, ";");
    if (!title.empty()) pl.addReplace(*new GText(title), nullptr);
  }

  auto& as = *new GAnimStep();
  as.list_.reserve(pl.size());

  for (auto& it : pl.list()) as.list_.push_back(it);
  steps_.push_back(&as);
  return *this;
}

GAnim& GAnim::operator<<(GStrings const& steps) {
  for (auto& it : steps) {
    operator<<(it);
  }
  return *this;
}

GAnimStep& GAnimStep::operator<<(GString const& str) {   // AnimStep!
  list_.push_back(new GText(str));
  return *this;
}

void GAnim::init() {
  if (steps_.empty()) {
    props_.list_.clear();
    return;
  }
  omodes_.initialized = 1;

  timeout_ = GTime((double)duration_ / (duration_ / GPrefs::animTimeout()));
  if (!timer_) timer_ = new GTimer;
  timer_->setDelay(timeout_);
  timer_->setRepetitions(Infinite);
  *timer_ << on::change / [this]{timeoutCB();};
  elapsed_ = 0;

  if (props_.empty() && !steps_.empty()) {
    for (auto& p : steps_[0]->list_) {
      if (p) props_ << p->clone(true);
    }
  }
}

bool GAnim::onAdd(Gadget* g) {
  if (!g) return true;

  if (startdelayed_) {start(); startdelayed_ = false;}

  if (props_.empty() && !steps_.empty()) {
    for (auto& p : steps_[0]->list_) {
      if (p) props_ << p->clone(true);
    }
  }
  for (auto& p : props_.list_) g->addProp(*p, false);
  return true;
}

void GAnim::onRemove(Gadget* g) {
  if (!g) return;
  stop();
  for (auto& p : props_.list_) g->removeProp(*p, false);
}

GAnim& GAnim::start(bool yes) {
  if (yes && !omodes_.initialized) {
    init();
    if (!omodes_.initialized) startdelayed_ = true;
  }
  if (!timer_ || props_.empty()) return *this;

  if (yes == isRunning()) return *this;

  timer_->start(false);

  if (yes) { // start
    timer_->set(0, true);  // true => forces timerCB() to be called
    elapsed_ = 0;
    timer_->start(true);
  }
  else { // stop
    // neeeded to update gadget after anim, must be done before
    // removeProp() (cant update anything after that)
    for (auto& p : props_.list_) {
      if (auto* m = p->toMutableProp()) m->update();
    }
    gprotect<GAnim> protect(this);  // needed be things done after callbacks
    GEvent e(GTrigger::done, GCore::elapsedTime());
    fireCallbacks(e);
    if (repeat_ == GTimer::OneShot) {
      // Kill the anim safely: adding and removing a reference will make
      // the anim to be destroyed only if there is no other ref pointing it
      gptr<GAnim> p = this;
    }
  }
  return *this;
}

void GAnim::restart() {
  if (!timer_ || props_.empty()) return;
  timer_->restart();
}

void GAnim::fireCallbacks(GEvent& e) {
  if (steps_.empty()) return;
  // avoids destructing object while firing callbacks
  gprotect<GAnim> protect(this);
  e.aux_ = this;
  calls_.fire(e);
}

void GAnim::apply(Gadget* g, GRender& r, Specif) {
  if (isRunning()) {
    for (auto& p : props_.list_) {
      if (p) p->apply(g, r, GStyle::AnimSpecif);
    }
  }
}

void GAnim::updateProps(GAnimStep& from, GAnimStep& to, float alpha) {
  size_t k{};
  //cerr << "anim step " <<endl;
  for (auto& it : props_.list_) {
    auto* p = it.get() ? it->toMutableProp() : nullptr;

    if (p && k < from.list_.size() && k < to.list_.size()
        && from.list_[k] && to.list_[k]) {
      /*
      cerr << "- " << it->typeName()
      << " / " << from.list_[k]->typeName()
      << " / " << to.list_[k]->typeName()
      << endl;
       */
      /* cerr <<"updateProps " <<  k << " / " << alpha
      << "     " << *from.list_[k]<< "    "<< *to.list_[k] << endl; */
      bool stat = p->blend(*from.list_[k], *to.list_[k], alpha);
      if (!stat) cerr << "blend failed" << endl;
    }
    k++;
  }
  GEvent e(GTrigger::changeaction, GCore::elapsedTime());
  fireCallbacks(e);
}

void GAnim::timeoutCB() {
  elapsed_ += timeout_;
  if (repeat_ > 0 && elapsed_ >= duration_ * repeat_) {
    //cerr << "GAnim::Done " << endl;
    stop();
    return;
  }
  double stepdur = (double)duration_ / (steps_.size()-1);
  size_t k = size_t(elapsed_ / stepdur);
  size_t kk = k % (steps_.size()-1);
  //cerr << "GAnim "<< k << " " << kk << endl;

  auto* from = steps_[kk];
  auto* to = (kk+1) < steps_.size() ? steps_[kk+1] : steps_[0];
  if (!from || !to) return;

  // time between two steps
  GTime now = GTime(elapsed_ - k * stepdur);
  float alpha = 1.f - float(now / stepdur);
  updateProps(*from, *to, alpha);
}

}
