//
// Prop that enforces exclusive selection in a Box
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Choice_hpp
#define Guit_Choice_hpp
#include <gint.hpp>
namespace guit {

/// Enforces exclusive selection in a Box.
/// When adedd to a box, this prop enforces exclusive selection on the checkable children
/// of the Box. It also makes its item/buttons children checkable even if not by default.
/// Any gagdget can be make checkable by calling its setCheckable() method.
/// @note GChoiceBox and subclasses rely on GChoice.
class GChoice : public GInt {
public:
  /// Constructor.
  /// _index_ is the index of the ckeched child. This is the position in the child list taking into account
  /// only checkable chidren. -1 means no child selected, it's the default value (no arg constr.).
  /// When provided, min and max are the mininum and maximun values (should be postive).
  /// @{
  GChoice();
  GChoice(int index);
  GChoice(int index, int min, int max);
  GChoice(GString const& index);
  GChoice(GChoice const&);
  /// @}

  GChoice* clone(bool copy_value) const override;
  GChoice* toChoice() override {return this;}

  /// Returns the selected (ie checked) item or its index.
  /// See GChoiceBox::chosenItem()  and GChoiceBox::chosenIndex()  for explanation.
  /// @{
  int chosenIndex() const {return value_;}
  Gadget* chosenItem(const GBox&) const;
  /// @}

  /// Changes the selected (ie checked) item.
  /// See GChoiceBox::choose() for explanation.
  /// @{
  virtual void choose(int index) {set(index,false);}
  virtual void choose(GBox&, Gadget* item);
  virtual void choose(GBox&, GString const& name);
  GChoice& operator=(int index) {set(index,false); return *this;}
  GChoice& operator=(GInt const& index) {set(index,false); return *this;}
  GChoice& operator=(GString const& index) {set(index); return *this;}
  /// @}

  /// Selects (ie checks) the next or previous checkable item relatively to the current selection.
  /// @{
  GChoice& operator++() {increase(+1); return *this;}
  GChoice& operator++(int) {increase(+1); return *this;}
  GChoice& operator--() {increase(-1); return *this;}
  GChoice& operator--(int) {increase(-1); return *this;}
  /// @}

  /// Selects (ie checks) a checkable item relatively to the current selection.
  /// @{
  GChoice& operator+=(int value) {increase(value); return *this;}
  GChoice& operator-=(int value) {increase(-value); return *this;}
  using GInt::increase;
  void increase(int const& value) override;
  /// @}

  using GType = GPropType_<GChoice,GInt>;
  static GType& Type();
  GType& type() const override {return Type();}

  /// @internal
  /// @{
  virtual void addBoxChild(GBox&, Gadget* child, int pos);
  void radioCB(GBox&);
  void addModes(GadgetModes&) override;
  void removeModes(GadgetModes&) override;
  /// @}
};

/// Returns a GChoice created in the heap.
/// @{
GChoice& Choice();
GChoice& Choice(int index);
GChoice& Choice(int index, int min, int max);
GChoice& Choice(GString const& index);
/// @}

}
#endif
