//
// Color and Background Props
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <unordered_map>
#include <sstream>
#include <gadget.hpp>
#include <gtext.hpp>
#include <gexpr.hpp>
#include <gcontrol.hpp>
#include <grender.hpp>
#include <gstring.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP_TYPE(Color, GPropType::Color)
GIMPLEMENT_PROP_TYPE(Background, GPropType::Background)

static GCore::BeginPlainObjects begin;

GColor
GColor::none(0.f, 0.f, 0.f, 0.f),
GColor::white(1.f, 1.f, 1.f),
GColor::black(0.f, 0.f, 0.f),
GColor::gray(0.5f, 0.5f, 0.5f),
GColor::lightgray(0.83f, 0.83f, 0.83f),
GColor::dimgray(0.4f, 0.4f, 0.4f),
GColor::red(1.f, 0.f, 0.f),
GColor::green(0.f, 0.5f, 0.f),   // from CSS=008000
GColor::blue(0.f, 0.f, 1.f),
GColor::lightblue(173.f/255.f, 216.f/255.f, 230.f/255.f),
GColor::guitblue(0.066f, 0.46f, 0.65f),
GColor::cyan(0.f, 1.f, 1.f),
GColor::magenta(1.f, 0.f, 1.f),
GColor::yellow(1.f, 1.f, 0.f),
GColor::orange(1.f, 0.5f, 0.f);

GBackground
GBackground::none(0.f, 0.f, 0.f, 0.f),
GBackground::white(1.f, 1.f, 1.f, 1.f),
GBackground::black(0.f, 0.f, 0.f, 1.f),
GBackground::gray(0.5f, 0.5f, 0.5f, 1.f),
GBackground::lightgray(0.83f, 0.83f, 0.83f, 1.f),
GBackground::dimgray(0.412f, 0.412f, 0.412f, 1.f),
GBackground::red(1.f, 0.f, 0.f, 1.f),
GBackground::green(0.f, 0.5f, 0.f, 1.f),   // from CSS=008000
GBackground::blue(0.f, 0.f, 1.f, 1.f),
GBackground::lightblue(173.f/255.f, 216.f/255.f, 230.f/255.f, 1.f),
GBackground::guitblue(17.f/255.f, 117.f/255.f, 217.f/255.f, 1.f),
GBackground::cyan(0.f, 1.f, 1.f, 1.f),
GBackground::magenta(1.f, 0.f, 1.f, 1.f),
GBackground::yellow(1.f, 1.f, 0.f, 1.f),
GBackground::orange(1.f, 0.5f, 0.f, 1.f);

static GCore::EndPlainObjects end;

GColor::GColor() = default;

GColor::GColor(float r, float g, float b, float a)
: Super(GRgba(r, g, b, a)) {}

GColor::GColor(const GRgba& c) : Super(c) {}

GColor::GColor(const GColor& c) : Super(c.value_) {}

GColor::GColor(const GColor& c, float a)
: Super(GRgba(c.value_.red, c.value_.green, c.value_.blue, a)) {}

GColor::GColor(GString const& name) {
  set(name);
}

GColor* GColor::clone(bool copy_value) const {
  return copy_value ? new GColor(value_) : new GColor;
}

bool GColor::isValid(GRgba const& c) const  {
  return
  c.red >= 0.f && c.red <= 1.f
  && c.green >= 0.f && c.green <= 1.f
  && c.blue >= 0.f && c.blue <= 1.f
  && c.alpha >= 0.f && c.alpha <= 1.f;
}

bool GColor::doChange(GRgba const& value) {
  auto v = value;
  if (v.red < 0.f)   v.red = 0.f;   else if (v.red > 1.f)   v.red = 1.f;
  if (v.green < 0.f) v.green = 0.f; else if (v.green > 1.f) v.green = 1.f;
  if (v.blue < 0.f)  v.blue = 0.f;  else if (v.blue > 1.f)  v.blue = 1.f;
  if (v.alpha < 0.f) v.alpha = 0.f; else if (v.alpha > 1.f) v.alpha = 1.f;
  if (v == value_ && pmodes_.isSet) return false;
  pmodes_.isSet = 1;
  value_ = v;
  return true;
}

GColor& GColor::operator=(GRgba const& v) {set(v,false); return *this;}
GColor& GColor::operator=(GString const& v) {set(v); return *this;}
GColor& GColor::operator=(const GColor& v) {set(v); return *this;}

void GColor::set(float r, float g, float b, float a) {
  set(GRgba(r, g, b, a),false);
}

void GColor::setRed(float val) {
  set(GRgba(val, value_.green, value_.blue, value_.alpha),false);
}

void GColor::setGreen(float val) {
  set(GRgba(value_.red, val, value_.blue, value_.alpha),false);
}

void GColor::setBlue(float val) {
  set(GRgba(value_.red, value_.green, val, value_.alpha),false);
}

void GColor::setAlpha(float val) {
  set(GRgba(value_.red, value_.green, value_.blue, val),false);
}

GColor& GColor::operator<<=(GTextExpr& expr) {
  bindImpl(expr);
  set(expr.stringValue());
  return *this;
}

void GColor::blend(GColor const& a, GColor const& b, float coeff) {
  set(GRgba::blend(a.value_, b.value_, coeff), false);
}

bool GColor::blend(GProp const& a, GProp const& b, float coeff) {
  GColor c1, c2;
  auto* p1 = dynamic_cast<const GColor*>(&a);
  if (!p1) {
    if (auto* s1 = a.toText()) {c1 = s1->value(); p1 = &c1;}
  }
  auto* p2 = dynamic_cast<const GColor*>(&b);
  if (!p2) {
    if (auto* s2 = b.toText()) {c2 = s2->value(); p2 = &c2;}
  }
  if (p1 && p2) {set(GRgba::blend(p1->value(), p2->value(), coeff), false); return true;}
  else return false;
}

void GColor::updateGadget(Gadget& g) {
  //cerr << "\n**********  GColor::updateGadget: " << value_ <<"\n"<< endl;
  g.update();
}

void GColor::apply(Gadget*, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Color]) return;
  r.specif[GPropType::Color] = specif;
  
  if (this == &GColor::none) r.color_ = nullptr;
  else if (value_.alpha >= 0) r.color_ = &value_;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GBackground::GBackground() = default;

GBackground::GBackground(float r, float g, float b, float a)
: GColor(r,g,b,a) {}

GBackground::GBackground(const GRgba& c) : GColor(c) {}

GBackground::GBackground(const GColor& c) : GColor(c) {}

GBackground::GBackground(const GColor& c, float a) : GColor(c,a) {}

GBackground::GBackground(GString const& name) {
  set(name);
}

GBackground* GBackground::clone(bool copy_value) const {
  return copy_value ? new GBackground(value_) : new GBackground;
}

void GBackground::apply(Gadget*, GRender& r, Specif specif) {
  if (specif < r.specif[GPropType::Background]) return;
  r.specif[GPropType::Background] = specif;

  if (this == &GBackground::none) r.background_ = nullptr;
  else if (value_.alpha >= 0) r.background_ = &value_;
}

GBackground& GBackground::operator=(GRgba const& v) {
  set(v,false); return *this;
}

GBackground& GBackground::operator=(const GColor& v){
  set(v.value(),false); return *this;
}

GBackground& GBackground::operator=(GString const& v) {
  set(v); return *this;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static float hexToValue(char a, char b) {
  int v1;
  if (a >= '0' && a <= '9') v1 = a - '0';
  else if (a >= 'a' && a <= 'f') v1 = a - 'a' + 10;
  else if (a >= 'A' && a <= 'F') v1 = a - 'A' + 10;
  else v1 = 0;
  int v2;
  if (b >= '0' && b <= '9') v2 = b - '0';
  else if (b >= 'a' && b <= 'f') v2 = b - 'a' + 10;
  else if (b >= 'A' && b <= 'F') v2 = b - 'A' + 10;
  else v2 = 0;
  return (v1*16+v2) / 255.f;
}

static void hexcodeToRgba(GRgba& color, GString const& code) {
  auto count = code.size();
  if (count < 7) gerror("GColor","format should be #RRGGBB or #RRGGBBAA: "+code);
  color.red   = count > 2 ? hexToValue(code[1],code[2]) : 0.f;
  color.green = count > 4 ? hexToValue(code[3],code[4]) : 0.f;
  color.blue  = count > 6 ? hexToValue(code[5],code[6]) : 0.f;
  color.alpha = count > 8 ? hexToValue(code[7],code[8]) : 1.f;
}

bool gconvert(GString const& from, GRgba& to) {
  return to.set(from);
}

bool gconvert(const GRgba& from, GString& to) {
  to = from.stringValue();
  return true;
}

GString GRgba::stringValue() const {
  std::ostringstream ss;
  ss.imbue(std::locale("C"));
  ss.precision(GPrefs::precision());
  ss << red << ' ' << green << ' ' << blue;
  if (alpha != 1.f) ss << ' '<< alpha;
  return ss.str();
}

GRgba::GRgba(float red, float green, float blue, float alpha)
: red(red), green(green), blue(blue), alpha(alpha) {}

GRgba::GRgba(GString const& colorspec) {set(colorspec);}

void GRgba::set(float red_, float green_, float blue_, float alpha_) {
  red = red_; green = green_; blue = blue_; alpha = alpha_;
}

// returns alpha * this_color + (1-alpha) * color2
GRgba GRgba::blend(const GRgba& c1, const GRgba& c2, float coeff) {
  GRgba col;
  float beta = 1.f - coeff;
  col.red   = coeff * c1.red   + beta * c2.red;
  col.green = coeff * c1.green + beta * c2.green;
  col.blue  = coeff * c1.blue  + beta * c2.blue;
  col.alpha = coeff * c1.alpha + beta * c2.alpha;
  return col;
}

bool GRgba::operator==(const GRgba& c) const {
  return red == c.red && green == c.green && blue == c.blue && alpha == c.alpha;
}

bool GRgba::operator!=(const GRgba& c) const {
  return red != c.red || green != c.green || blue != c.blue || alpha != c.alpha;
}

std::ostream& operator<<(std::ostream& out, const GRgba& c) {
  out << c.red << " "<< c.green << " "<< c.blue << " "<< c.alpha;
  return out;
}

std::istream& operator>>(std::istream& in, GRgba& c) {
  in >> c.red >> c.green >> c.blue >> c.alpha;
  return in;
}

static const char* findColorName(GString const& name);

bool GRgba::set(GString const& colorname) {
  red = green = blue = 0.f;
  alpha = 1.f;
  
  GString name = strlower(colorname);
  strtrim(name);
  if (name.empty()) {
    return false;
  }
  if (name[0] == '#') {
    hexcodeToRgba(*this, name);
    return true;
  }
  if (!std::isdigit(name[0])) {
    if (const char* c = findColorName(name)) {
      hexcodeToRgba(*this, c);
      return true;
    }
    else {
      gerror("GColor","color name not found: '" + name +"'");
      //alpha = 0.f;  // !!
      return false;
    }
  }
  std::istringstream ss(name.data());
  ss.imbue(std::locale("C"));
  ss >> red >> green >> blue;
  if (ss.fail()) {
    gerror("GColor","color can have 3 or 4 components: " + name);
    return false;
  }
  ss >> alpha;
  return true;
  /*
  GString curloc = setlocale(LC_ALL,NULL);
  setlocale(LC_NUMERIC, "C");
  auto n = sscanf(name.data(), "%f %f %f %f", &red, &green, &blue, &alpha);
  if (n < 3 || n > 4)
    gerror("GColor","color can have 3 or 4 components: " + name);
  setlocale(LC_NUMERIC, curloc.data());
   return n >= 3;
   */
}

static const char* findColorName(GString const& name) {
  static const std::map<GString, const char*> colorNames{
    {"none", "#00000000"},         // specific to the toolkit
    {"aliceblue", "#f0f8ff"},
    {"antiquewhite", "#faebd7"},
    {"aqua", "#00ffff"},
    {"aquamarine", "#7fffd4"},
    {"azure", " #f0ffff"},
    {"beige", "#f5f5dc"},
    {"bisque", "#ffe4c4"},
    {"black", "#000000"},
    {"blanchedalmond", "#ffebcd"},
    {"blue", "#0000ff"},
    {"blueviolet", "#8a2be2"},
    {"brown", "#a52a2a"},
    {"burlywood", "#deb887"},
    {"cadetblue", "#5f9ea0"},
    {"chartreuse", "#7fff00"},
    {"chocolate", "#d2691e"},
    {"coral", "#ff7f50"},
    {"cornflowerblue", "#6495ed"},
    {"cornsilk", "#fff8dc"},
    {"crimson", "#dc143c"},
    {"cyan", "#00ffff"},
    {"darkblue", "#00008b"},
    {"darkcyan", "#008b8b"},
    {"darkgoldenrod", "#b8860b"},
    {"darkgray", "#a9a9a9"},
    {"darkgrey", "#a9a9a9"},
    {"darkgreen", "#006400"},
    {"darkkhaki", "#bdb76b"},
    {"darkmagenta", "#8b008b"},
    {"darkolivegreen", "#556b2f"},
    {"darkorange", "#ff8c00"},
    {"darkorchid", "#9932cc"},
    {"darkred", "#8b0000"},
    {"darksalmon", "#e9967a"},
    {"darkseagreen", "#8fbc8f"},
    {"darkslateblue", "#483d8b"},
    {"darkslategray", "#2f4f4f"},
    {"darkslategrey", "#2f4f4f"},
    {"darkturquoise", "#00ced1"},
    {"darkviolet", "#9400d3"},
    {"deeppink", "#ff1493"},
    {"deepskyblue", "#00bfff"},
    {"dimgray", "#696969"},
    {"dimgrey", "#696969"},
    {"dodgerblue", "#1e90ff"},
    {"firebrick", "#b22222"},
    {"floralwhite", "#fffaf0"},
    {"forestgreen", "#228b22"},
    {"fuchsia", "#ff00ff"},
    {"gainsboro", "#dcdcdc"},
    {"ghostwhite", "#f8f8ff"},
    {"gold", "#ffd700"},
    {"goldenrod", "#daa520"},
    {"gray", "#808080"},
    {"grey", "#808080"},
    {"green", "#008000"},
    {"greenyellow", "#adff2f"},
    {"guitblue", "#1175d9"}, // 17, 117, 217
    {"honeydew", "#f0fff0"},
    {"hotpink", "#ff69b4"},
    {"indianred", "#cd5c5c"},
    {"indigo", "#4b0082"},
    {"ivory", "#fffff0"},
    {"khaki", "#f0e68c"},
    {"lavender", "#e6e6fa"},
    {"lavenderblush", "#fff0f5"},
    {"lawngreen", "#7cfc00"},
    {"lemonchiffon", "#fffacd"},
    {"lightblue", "#add8e6"},
    {"lightcoral", "#f08080"},
    {"lightcyan", "#e0ffff"},
    {"lightgoldenrodyellow", "#fafad2"},
    {"lightgray", "#d3d3d3"},
    {"lightgrey", "#d3d3d3"},
    {"lightgreen", "#90ee90"},
    {"lightpink", "#ffb6c1"},
    {"lightsalmon", "#ffa07a"},
    {"lightseagreen", "#20b2aa"},
    {"lightskyblue", "#87cefa"},
    {"lightslategray", "#778899"},
    {"lightslategrey", "#778899"},
    {"lightsteelblue", "#b0c4de"},
    {"lightyellow", "#ffffe0"},
    {"lime", "#00ff00"},
    {"limegreen", "#32cd32"},
    {"linen", "#faf0e6"},
    {"magenta", "#ff00ff"},
    {"maroon", "#800000"},
    {"mediumaquamarine", "#66cdaa"},
    {"mediumblue", "#0000cd"},
    {"mediumorchid", "#ba55d3"},
    {"mediumpurple", "#9370db"},
    {"mediumseagreen", "#3cb371"},
    {"mediumslateblue", "#7b68ee"},
    {"mediumspringgreen", "#00fa9a"},
    {"mediumturquoise", "#48d1cc"},
    {"mediumvioletred", "#c71585"},
    {"midnightblue", "#191970"},
    {"mintcream", "#f5fffa"},
    {"mistyrose", "#ffe4e1"},
    {"moccasin", "#ffe4b5"},
    {"navajowhite", "#ffdead"},
    {"navy", "#000080"},
    {"oldlace", "#fdf5e6"},
    {"olive", "#808000"},
    {"olivedrab", "#6b8e23"},
    {"orange", "#ffa500"},
    {"orangered", "#ff4500"},
    {"orchid", "#da70d6"},
    {"palegoldenrod", "#eee8aa"},
    {"palegreen", "#98fb98"},
    {"paleturquoise", "#afeeee"},
    {"palevioletred", "#db7093"},
    {"papayawhip", "#ffefd5"},
    {"peachpuff", "#ffdab9"},
    {"peru", "#cd853f"},
    {"pink", "#ffc0cb"},
    {"plum", "#dda0dd"},
    {"powderblue", "#b0e0e6"},
    {"purple", "#800080"},
    {"rebeccapurple", "#663399"},
    {"red", "#ff0000"},
    {"rosybrown", "#bc8f8f"},
    {"royalblue", "#4169e1"},
    {"saddlebrown", "#8b4513"},
    {"salmon", "#fa8072"},
    {"sandybrown", "#f4a460"},
    {"seagreen", "#2e8b57"},
    {"seashell", "#fff5ee"},
    {"sienna", "#a0522d"},
    {"silver", "#c0c0c0"},
    {"skyblue", "#87ceeb"},
    {"slateblue", "#6a5acd"},
    {"slategray", "#708090"},
    {"slategrey", "#708090"},
    {"snow", "#fffafa"},
    {"springgreen", "#00ff7f"},
    {"steelblue", "#4682b4"},
    {"tan", "#d2b48c"},
    {"teal", "#008080"},
    {"thistle", "#d8bfd8"},
    {"tomato", "#ff6347"},
    {"turquoise", "#40e0d0"},
    {"violet", "#ee82ee"},
    {"wheat", "#f5deb3"},
    {"white", "#ffffff"},
    {"whitesmoke", "#f5f5f5"},
    {"yellow", "#ffff00"},
    {"yellowgreen", "#9acd32"}
  };
  auto it = colorNames.find(name);
  if (it == colorNames.end()) return nullptr;
  else return it->second;
}
}
