//
// Modes of a gadget, control whether a gadget is enabled, shown, checked, etc.
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gmodes.hpp>
#include <gbox.hpp>
#include <gcontrol.hpp>
#include <gstring.hpp>
#include <ggraphics.hpp>
#include <grender.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP_TYPE(Modes, GPropType::Modes)
GIMPLEMENT_PROP_TYPE(Enable,GPropType::Auto)
GIMPLEMENT_PROP_TYPE(Show,GPropType::Auto)
GIMPLEMENT_PROP_TYPE(Check,GPropType::Auto)

GEnable::GEnable() = default;
GEnable::GEnable(bool value) : GBool(value) {}
GEnable::GEnable(GString const& value) : GBool(value) {}

GEnable& GEnable::operator=(bool value) {set(value,false); return *this;}
GEnable& GEnable::operator=(GBool const& p) {set(p.value(),false); return *this;}
GEnable& GEnable::operator=(GString const& value) {set(value); return *this;}

GEnable* GEnable::clone(bool copy_value) const {
  return copy_value ? new GEnable(value_) : new GEnable;
}

void GEnable::updateGadget(Gadget& g) {
  g.enable(value());
}

void GEnable::apply(Gadget* g, GRender& r, Specif specif) {
  if (g) g->enable(value());
}


GShow::GShow() = default;
GShow::GShow(bool value) : GBool(value) {}
GShow::GShow(GString const& value) : GBool(value) {}

GShow& GShow::operator=(bool value) {set(value,false); return *this;}
GShow& GShow::operator=(GBool const& p) {set(p.value(),false); return *this;}
GShow& GShow::operator=(GString const& value) {set(value); return *this;}

GShow* GShow::clone(bool copy_value) const {
  return copy_value ? new GShow(value_) : new GShow;
}

void GShow::updateGadget(Gadget& g) {
  g.show(value());
}

void GShow::apply(Gadget* g, GRender& r, Specif specif) {
  if (g) g->show(value());
}


GCheck::GCheck() = default;
GCheck::GCheck(bool value) : GBool(value) {}
GCheck::GCheck(GString const& value) : GBool(value) {}

GCheck& GCheck::operator=(bool value) {set(value,false); return *this;}
GCheck& GCheck::operator=(GBool const& p) {set(p.value(),false); return *this;}
GCheck& GCheck::operator=(GString const& value) {set(value); return *this;}

GCheck* GCheck::clone(bool copy_value) const {
  return copy_value ? new GCheck(value_) : new GCheck;
}

void GCheck::addModes(GadgetModes& m) {
  m.hasCheck = 1;
}

void GCheck::removeModes(GadgetModes& m) {
  m.hasCheck = 0;
}

void GCheck::apply(Gadget* g, GRender& r, Specif specif) {
  if (g) g->selectImpl(value(), Gadget::FromSelect);
}

void GCheck::updateGadget(Gadget& g) {
  g.modes_.activable = true;
  g.modes_.checkable = true;    // OK pour stack?
  g.modes_.hasCheck = true;
  g.selectImpl(value(), Gadget::FromSelect);
}


GModes::GModes(GString const& s) {
  std::vector<GString> tab;
  strsplit(s, tab, ",");

  for (auto& it : tab) {
    bool val = 1;
    if (it.empty()) continue;
    else if (it[0] == '-' || it[0] == '!') {val = 0; it.erase(0,1);}
    else if (it[0] == '+') {val = 1; it.erase(0,1);}

    uint64_t type = strid(strlower(it));
    modes_.push_back(Mode{type, val});      // TODO: verifier que le mode exsite !!!
  }
}

void GModes::setModes(GadgetModes& m) {
  //cerr << "GModes::setModes" << endl;
  for (auto& it : modes_) {
    switch (it.type) {
      case strid("checked"):
        m.checked = it.val;
        break;
      case strid("checkable"):
        m.checkable = it.val;
        break;
      case strid("focusable"):
        m.focusable = it.val;
        break;
        //case strid("keepactive"):
        //m.keepActive = it.val;
        break;
      case strid("activable"):
        m.activable = it.val;
        break;
      case strid("downactivable"):
        m.downActivable = it.val;
        if (it.val) m.activable = it.val;
        break;
      case strid("catchevents"):
        m.catchEvents = it.val;
        break;
      case strid("shown"):
        m.shown = it.val;
        break;
      case strid("enabled"):
        m.enabled = it.val;
        break;
      case strid("richtext"):
        m.richText = it.val;
        break;
      case strid("multiline"):
        m.multiLine = it.val;
        break;
      case strid("stealtitle"):
        m.stealTitle = it.val;
        break;
      case strid("choicenone"):
        m.choiceBehavior = uint8_t(GBox::ChoiceBehavior::NoChoice);
        break;
      case strid("choiceselect"):
        m.choiceBehavior = uint8_t(GBox::ChoiceBehavior::Select);
        break;
      case strid("choiceshow"):
        m.choiceBehavior = uint8_t(GBox::ChoiceBehavior::Show);
        break;
      case strid("menulayout"):
        m.menuLayout = it.val;
        break;
      case strid("internal"):
        m.internal = it.val;
        break;
    }
  }
}

void GModes::addModes(GadgetModes& m) {
  setModes(m);
}

void GModes::apply(Gadget* g, GRender& r, Specif specif) {
  if (!g || g->settings_.modesSet) return;
  if (g) {
    g->settings_.modesSet = 1;
    setModes(g->modes_);
  }
}

}
