//
// Base clas for Number props
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef Guit_Number_hpp
#define Guit_Number_hpp
#include <limits>
#include <gvarprop.hpp>
namespace guit {

class GNumber : public GVarProp {
public:
  using IsNumber = bool;

  bool isNumber() const override {return true;}
  GNumber* toNumber() override {return this;}

  virtual void setF(float value, bool callalways = false) = 0;
  void set(GString const& value) override = 0;
  bool set(GProp const& value) override = 0;

  virtual void increaseF(float value) = 0;
  virtual void increase() = 0;
  virtual void decrease() = 0;

  bool blend(GProp const& a, GProp const& b, float mix) override;

  /// Changes/returns the range.
  /// @{
  virtual GNumber& setRangeF(float min, float max) = 0;
  virtual float minF() const = 0;
  virtual float maxF() const = 0;
  /// @}

  /// Returns the value scaled in a [0.f, 1.f] interval.
  /// Makes sense only if the range is set (see setRangeF()).
  virtual double normedValue() const = 0;

  /// Returns the lowest/highest possible values for this type.
  /// @{
  virtual float lowestF() const = 0;
  virtual float highestF() const = 0;
  /// @}

  void updateGadget(Gadget&) override;

protected:
  GNumber() = default;
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template <class T> class GNumber_ : public GNumber {
public:
  using ValueType = T;

  /// Returns the value.
  /// @{
  T const& value() const {return value_;}
  T const& operator()() const {return value_;}
  operator T const&() const {return value_;}
  /// @}

  /// Returns the value converted to a different type.
  /// @{
  bool  boolValue()  const override {return bool(value_);}
  float floatValue() const override {return float(value_);}
  GString stringValue() const override {return guit::toString(value_);}
  /// @}

  void setF(float value, bool callalways = false) override {
    set(static_cast<T>(value), callalways);
  }

  virtual void set(T const& value, bool callalways = false) {
    if (omodes_.immutable) immutableError("set");
    else if (doChange(value) || callalways) propagateChange();
  }

  void set(GString const& value) override {
    T val{}; gconvert(value, val); set(val, false);
  }

  bool set(GProp const& prop) override {
    set(static_cast<T>(prop.floatValue()), false);
    return true;
  }

  GNumber_<T>& setRangeF(float min, float max) override {
    return setRange(static_cast<T>(min), static_cast<T>(max));
  }

  GNumber_<T>& setRange(T const& min, T const& max) {
    //if (max < min) max = min;
    min_ = min;
    max_ = max;
    set(value_,false);  // pas bon car value_ == value_ donc fait rien !!
    return *this;
  }

  using GNumber::increase;
  void increaseF(float delta) override {increase(static_cast<T>(delta));}
  virtual void increase(T const& delta) {set(value_ + delta);}

  double normedValue() const override {
    return (max_ == min_) ? 0. : (double)(value_ - min_) / (max_ - min_);
  }

  /// Returns the range (see setRange()).
  /// @{
  T min() const {return min_;}
  T max() const {return max_;}
  float minF() const override {return float(min_);}
  float maxF() const override {return float(max_);}
  /// @}

  /// Returns the lowest/highest possible values.
  /// @{
  static T lowest() {return std::numeric_limits<T>::lowest();}
  static T highest() {return std::numeric_limits<T>::max();}
  float lowestF() const override {return float(lowest());}
  float highestF() const override {return float(highest());}
  /// @}

  /// @internal
  /// @{
  virtual bool isValid(T const& val) const {
    return (val >= min_ && val <= max_);
  }

  virtual bool doChange(T const& v) {
    if (v != value_ || !pmodes_.isSet) {
      pmodes_.isSet = 1;
      if (v > max_) value_ = max_;
      else if (v < min_) value_ = min_;
      else value_ = v;
      return true;
    }
    else return false;
  }

  T value_{};
  T min_{lowest()};
  T max_{highest()};
  /// @}
};

}
#endif
