//
// List of Props
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <gproplist.hpp>
#include <gevent.hpp>
#include <gexpr.hpp>
#include <gnotifier.hpp>
#include <gcontrol.hpp>
#include <gstyle.hpp>
using namespace std;
namespace guit {

GPropList::GPropList(std::initializer_list<gptr<GProp>> list)
: list_(list) {}

size_t GPropList::size() const {
  size_t count = 0;
  for (auto it = list_.begin(); it != list_.end(); ++it) count++;
  return count;
}

void GPropList::fire(GEvent& e) {
  for (auto& it : list_)
    it->fire(e);
}

void GPropList::fire(GObject* aux) {
  GEvent e(GTrigger::changeaction, GCore::elapsedTime());
  e.aux_ = aux;
  for (auto& it : list_)
    it->fire(e);
}

void GPropList::addBegin(GPropList const& pl) {
  auto pos = list_.before_begin();
  for (auto& from : pl.list_) {
    from->onAdd(nullptr);
    pos = list_.insert_after(pos, from.get());
  }
}

void GPropList::addEnd(GPropList const& pl) {
  auto pos = list_.before_begin();
  for (auto it = list_.begin(); it != list_.end(); ++it) pos = it;
  for (auto& from : pl.list_) {
    from->onAdd(nullptr);
    pos = list_.insert_after(pos, from.get());
  }
}

void GPropList::addBegin(GProp& p) {   // see push_front()
  auto pos = list_.before_begin();
  p.onAdd(nullptr);
  list_.insert_after(pos, p);
}

void GPropList::addEnd(GProp& p) {
  auto pos = list_.before_begin();
  for (auto it = list_.begin(); it != list_.end(); ++it) pos = it;
  p.onAdd(nullptr);
  list_.insert_after(pos, p);
}

void GPropList::addReplace(GProp& p, Gadget* g) {
  bool found = false;
  GPropList::iterator it, equiv_it;
  GPropList::iterator prev_it = list_.before_begin();

  // search prop with same role, if found will replace it or update it
  for (it = list_.begin(); it != list_.end(); ++it) {
    if ((*it).get() == &p) {    // The prop already belongs to the list
      p.onAdd(g);
      return;
    }
    else if ((*it)->isEquivalent(p)) {
      found = true;
      equiv_it = it;
      break;
    }
    prev_it = it;
  }

  // the prop must not be added if onAdd() returns false
  // ???? est-ce qu'il faut faire cela si equiv_it est un model ????
  if (!p.onAdd(g)) return;

  if (!found) {
    // if not found add to the end of the list
    list_.insert_after(prev_it, p);
  }
  else {
    /*
     // ca remplace soit la prop, soit sa value, trop ambigu !
     auto* m = (*equiv_it)->toMutable(true);
     if (m && !p.isMutable()) m->set(p); else *equiv_it = p;
     */
    /*
    auto* m = (*equiv_it)->toMutableProp();
    if (m && !p.isMutable()) {
      cerr <<"WARNING: << behavior changed: "
      << m->typeName() << " = " << p.typeName() << endl;
    }*/

    // otherwise the prop is replaced in the proplist
    *equiv_it = p;
  }

  if (g && (g->eventMask() & GEventType::AddRemove) != 0) {
    GEvent e(on::add, GCore::elapsedTime());
    e.gadget_ = g;
    e.aux_ = &p;
    fire(e);
  }
}

/*
 GProp* GPropList::at(int index) {
 int k = 0;
 for (auto& it : list_) {
 if (k++ == index) return it;
 }
 return nullptr;
 }*/

GProp* GPropList::findProp(std::function<bool(GProp&)>& pred) const {
  for (auto& it : list_) {
    if ((pred)(*it)) return it;
  }
  return nullptr;
}

GProp* GPropList::findProp(GPropType& type) const {
  for (auto& it : list_) {
    auto& t = it->type();

    if (t.role_ == type.role_) {
      if (t == type) return it;
      else {
        auto* tpar = t.parent_;
        while (tpar) {
          if (*tpar == type) return it;
          tpar = t.parent_;
        }
      }
      return nullptr;
    }
  }
  return nullptr;
}

bool GPropList::remove(Gadget* g, GProp& prop) {
  return remove(g, [&prop](GProp& p){return &p == &prop;});
}

bool GPropList::remove(Gadget* g, GPropType& type) {
  return remove(g, [&type](GProp& p){
    return p.role() == type.role();
  });
}

bool GPropList::remove(Gadget* g, std::function<bool(GProp&)> pred) {
  auto prev_it = list_.before_begin();
  auto it = list_.begin();

  for ( ; it != list_.end(); ++it) {
    if ((pred)(*it->get())) break;
    prev_it = it;
  }
  if (it == list_.end()) return false;

  if (g) {
    if (auto* m = (*it)->toMutableProp()) m->stopNotify(*g);
  }
  it->get()->onRemove(g);

  if (g && (g->eventMask() & GEventType::AddRemove) != 0) {
    GEvent e(GTrigger::remove, GCore::elapsedTime());
    e.gadget_ = g;
    e.aux_ = it->get();
    fire(e);
  }
  list_.erase_after(prev_it);
  return true;
}

void GPropList::removeAll(Gadget* g) {
  for (auto& it : list_) {
    if (g) {
      if (auto* m = it->toMutableProp()) m->stopNotify(*g);
    }

    it->onRemove(g);

    if (g && (g->eventMask() & GEventType::AddRemove) != 0) {
      GEvent e(GTrigger::remove, GCore::elapsedTime());
      e.gadget_ = g;
      e.aux_ = it;
      fire(e);
    }
  }
  list_.clear();
}

GNotifier* GPropList::findNotifier(void* obj) {
  for (auto& it : list_) {
    if (auto* n = it->toNotifier()) {
      if (n->receiver() == obj) return n;
    }
  }
  return nullptr;
}

void GPropList::removeNotifiers(void* obj) {
  auto prev_it = list_.before_begin();
  auto it = list_.begin();

  // remove all occurences of notifiers pointing to obj
  while (it != list_.end()) {
    if (auto* n = (*it)->toNotifier()) {
      if (n->receiver() == obj) {
        (*it)->onRemove(nullptr);
        ++it;
        list_.erase_after(prev_it);
        continue;
      }
    }
    prev_it = it++;
  }
}

}
