//
// States, Transitions and Statecharts
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <algorithm>
#include <iostream>
#include <ganim.hpp>
#include <gtrigger.hpp>
#include <gevent.hpp>
#include <gstatechart.hpp>
#include <gtext.hpp>
#include <gchoicebox.hpp>
#include <gexpr.hpp>
#include <gcontrol.hpp>
using namespace std;
namespace guit {

GIMPLEMENT_PROP_TYPE(State, GPropType::States)
GIMPLEMENT_PROP_TYPE(SuperState, GPropType::States)
GIMPLEMENT_PROP_TYPE(StateChart, GPropType::States)

GState& State(GSuperState& par, GString const& name) {return *new GState(par, name);}
GSuperState& SuperState(GSuperState& par, GString const& name) {return *new GSuperState(par, name);}
GStateChart& StateChart(GString const& name) {return *new GStateChart(name);}


GState::~GState() {
  removeAll();  // removes callbacks
  removeTransitions();
  delete name_;
}

GState::GState(GString const& name) {
  if (!name.empty()) name_ = new GString(name);
}

GState::GState(GSuperState& s, GString const& name) : GState(name) {
  s << this;
}

GString const GState::name() const {
  return name_ ? *name_ : "";
}

bool GState::entered() const {
  return parent_ && parent_->currentState() == this;
}

GString GState::stringValue() const {
  return entered() ? "entered" : "exited";
}

GStateTransition& GState::operator>>(GState& state2) {
  return *new GStateTransition(*this, state2);
}

GStateTransition& GState::operator>>(GState* state2) {
  if (!state2) {
    error("operator>>","null target state");
    return *new GStateTransition;
  }
  return *new GStateTransition(*this, *state2);
}

void GState::add(GStateTransition* t) {
  if (!t) {error("add", "Can't add null transition"); return;}
  t->state1_ = this;                      // d'abord enleve si t.state1 existe !!!
  
  auto it = std::find(transitions_.begin(), transitions_.end(), *t);
  if (it != transitions_.end())
    error("add", "Already a transition between these two states");
  else transitions_.push_back(t);
}

void GState::remove(GStateTransition* t) {
  if (!t) {
    error("remove", "Can't remove null transition");
    return;
  }
  for (auto it = transitions_.begin(); it != transitions_.end(); ++it) {
    if (it->get() == t) {
      transitions_.erase(it);
      return;
    }
  }
}

void GState::removeTransitions() {
  for (auto& it : transitions_) it->state1_ = nullptr;
  transitions_.clear();
}

void GState::enableGadgets(bool enable) {
  //cerr << "* enableGadgets: " << name() << " " << int(enable) << endl;
  for (auto* it : gadgets_) {
    GItem* item = it ? it->toItem() : nullptr;
    if (item) {
      // cerr << " - gadget: " << it->text () << " " << it->typeName()<< " "<< int(enable) << endl;
      it->enable(enable);
      it->update();
    }
  }
}

void GSuperState::enableGadgets(bool enable) {
  //cerr << "** enableGadgets: " << name() << " " << int(enable) << endl;
  GState::enableGadgets( enable);
  //cerr << " -- initial: "  << endl;
  if (initial_) initial_->enableGadgets(enable);
}

void GState::fireState(bool state) {
  GTrigger cond(GEventType::Change | GEventType::Action, state);
  GEvent e(cond, GCore::elapsedTime());
  e.aux_ = this;
  props_.fire(e);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GSuperState::~GSuperState() {
  removeStates();
}

GSuperState::GSuperState(GString const& name) : GState(name) {}

GSuperState::GSuperState(GSuperState& s, GString const& name) : GState(s, name) {}

void GSuperState::setInitial(GState& s) {
  setInitial(&s);
}

void GSuperState::setInitial(GState* s) {
  if (!s) initial_ = nullptr;
  else if (s->parent_ == this) initial_ = s;
  else error("setInitial","State argument should be a direct child of the HState");
}

void GSuperState::setCurrent(GState* s) {
  //cerr << "setcurrent "<< s << " " << states_.size() << endl;
  if (current_) {
    if (current_ == s) {
      current_->fireState(false);
      current_->fireState(true);
      return;
    }
    else {
      //cerr << "setcurrent fire false "<< current_ << endl;
      GState* ex = current_;
      current_ = nullptr;  // so that ex wont be entered when fire() is called
      ex->fireState(false);
    }
  }
  
  if (!s) {
    error("setCurrent", "Null state ");
    setCurrent(initial_);
  }
  else if (s->parent_ != this) {
    //error("setCurrent", "Wrong parent of state " + s->name());
    //cerr << "parent " << s->parent_->name()<< " - " << this->name() << endl;

    // on sort du HState => le remettre à initial
    // !!! il faudrait faire setCurrenut sur le superstate   !!!!!!!
    setCurrent(initial_);
  }
  else {
    // on reste dans le HState
    current_ = s;
    //cerr << "setcurrent fire true "<< current_ << endl;
    current_->fireState(true);
  }
}

void GSuperState::fireState(bool state) {
  //cerr << "GSuperState::fireState "<< current_ <<" "<< state << endl;
  GTrigger cond(GEventType::Change | GEventType::Action, state);
  GEvent e(cond, GCore::elapsedTime());
  e.aux_ = this;
  props_.fire(e);
  if (state) setCurrent(initial_);
}

GSuperState& GSuperState::operator<<(GState* s) {
  if (!s) error("addState", "State argument is null");
  else {
    s->parent_ = this;
    states_.push_back(s);
  }
  return *this;
}

GState& GSuperState::addState(GString const& name) {
  return *new GState(*this, name);
}

GSuperState& GSuperState::operator<<(GState& s) {
  return operator<<(&s);
}

void GSuperState::removeStates() {
  for (auto& it : states_) it->parent_ = nullptr;
  states_.clear();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GStateChart::GStateChart(GString const& name)
: GSuperState(name) {}

void GStateChart::run(bool yes) {
  if (!initial_) {
    error("run","No initial state");
    return;
  }
  if (yes) setCurrent(initial_); else setCurrent(nullptr);
}

void GStateChart::start(GState& init_state) {
  setInitial(init_state);
  run(true);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

class GCondTrans : public GCond {
public:
  GCondTrans(GTrigger&, GStateTransition&);
  void fire(GEvent&) override;
  void activate2(GEvent& e, bool state) override;
  bool onAdd(Gadget*) override;
  void onRemove(Gadget*) override;
  gptr<GStateTransition> trans_;
};

GStateTransition::GStateTransition() :
state1_(nullptr), state2_(nullptr), condtrans_(nullptr) {
}

GStateTransition::GStateTransition(GState& state1, GState& state2) :
state1_(&state1), state2_(&state2), condtrans_(nullptr) {
  state1.add(this);
}

GStateTransition::~GStateTransition() {
  // cerr << "~transition "<< this <<" "<<endl;
}

bool GStateTransition::canFire() const {
  //cerr << "canFire: " << state1_->parent()->currentState() <<" "<< state1_<< endl;
  return (state1_
          && state2_
          && state1_->parent()
          && state1_->parent()->currentState() == state1_);
}

void GStateTransition::doFire(GEvent& e) {
  // NOTE:
  // - the transition must not be performed now because, if there are several
  //   transitions they will be performed in sequence!
  // - postfire solves this problem: instead of doing the transition now,
  //   this will be done after attempting to fire all the props of proplist
  //   of the gadget with this event. As other transitions wont match (*)
  //   we are sure only this transition will be performed
  // (*) except of course if transitions are incorrect, e.g. if there are two
  // transitions starting from the same state with the same trigger)
  
  //cerr << "doFire: " << state1_ << " " <<  state2_ << endl;
  e.postFire([this](GEvent& e){
    //cerr << "-- OK transition::doFire: " << this << " " <<  state2_ << endl;
    if (condtrans_ && condtrans_->next()) condtrans_->next()->activate(e, true);
    state1_->enableGadgets(false);
    state1_->parent()->setCurrent(state2_);
    state2_->enableGadgets(true);
  });
}

GCondTrans::GCondTrans(GTrigger& trigger, GStateTransition& trans)
: GCond(trigger), trans_(trans) {
  trans.condtrans_ = this;
}

void GCondTrans::fire(GEvent& e) {
  //cerr << "condtrans::call " << endl;
  if (e.matches(trigger_,-1) > 0 && trans_ && trans_->canFire()) {
    //cerr << "-- OK condtrans::call " << endl;
    trans_->doFire(e);
  }
}

void GCondTrans::activate2(GEvent& e, bool state) {
  //cerr << "condtrans::activate "<< state << endl;
  if (state && trans_ && trans_->canFire()) {
    //cerr << "-- OK condtrans::activate " << endl;
    trans_->doFire(e);
  }
}

bool GCondTrans::onAdd(Gadget* g) {
  if (g && trans_ && trans_->state1_) trans_->state1_->gadgets_.push_front(g);
  return GCond::onAdd(g);
}

void GCondTrans::onRemove(Gadget* g) {
  if (g && trans_ && trans_->state1_) trans_->state1_->gadgets_.remove(g);
  GCond::onRemove(g);
}

// - - - - - -

void Gadget::add(GStateTransition& trans) {
  add(*new GCondTrans(GTrigger::action, trans));
}

GTimer& GTimer::operator<<(GStateTransition& trans) {
  add(*new GCondTrans(GTrigger::action, trans));
  return *this;
}

GCond& operator/(GTrigger& t, GStateTransition& trans) {
  return *new GCondTrans(t, trans);
}

GCond& operator/(GBoolExpr& e, GStateTransition& trans) {
  return *new GCondTrans(*new GExprTrigger(e), trans);
}

GCond& operator/(GCond& c, GStateTransition& trans) {
  c.addNext(new GCondTrans(c.trigger(), trans));
  return c;
}

GCond& operator/(GStateTransition& trans, std::function<void()> const& fun) {
  auto& c = *new GCondTrans(GTrigger::action, trans);
  c.addNext(& (c.trigger() / fun));
  return c;
}

GCond& operator/(GStateTransition& trans, GString const& text) {
  auto& c = *new GCondTrans(GTrigger::action, trans);
  c.addNext(new GCondProp(c.trigger(), *new GText(text)));
  return c;
}

GCond& operator/(GStateTransition& trans, GProp& prop) {
  auto& c = *new GCondTrans(GTrigger::action, trans);
  c.addNext(new GCondProp(c.trigger(), prop));
  return c;
}

GCond& operator/(GStateTransition& trans, GProp* prop) {
  auto& c = *new GCondTrans(GTrigger::action, trans);
  if (prop) c.addNext(new GCondProp(c.trigger(), *prop));
  else gerror("operator/", "argument can't be a null prop");
  return c;
}

}
