//
// internal wrappers
//
// Guit GUI Toolkit (https://www.telecom-paris.fr/~elc/guit/)
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com) - Telecom Paris
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
#ifndef Guit_PropWrapper_hpp
#define Guit_PropWrapper_hpp
#include <gprop.hpp>
#include <gtrigger.hpp>
namespace guit {

class GPropWrapper : public GProp {
public:
  GPropWrapper(GProp& p) : prop_(p) {}
  GPropWrapper* toPropWrapper() override {return this;}
  GProp* prop() {return prop_;}
  // either stringValue() or floatValue() MUST be redefined in subclasses!
  GString stringValue() const override;
  float floatValue() const override;
  gptr<GProp> prop_;
};

template <class T, class F>
class GFieldWrapper : public GPropWrapper {
  F& field_;
public:
  GFieldWrapper(T& prop, F& field) : GPropWrapper(prop), field_(field) {}
  GString stringValue() const override {return *field_;}
};

template <class T, class F>
class GNumFieldWrapper : public GPropWrapper {
  F& field_;
public:
  GNumFieldWrapper(T& prop, F& field) : GPropWrapper(prop), field_(field) {}
  float floatValue() const override {return field_;}
};

template <class T>
class GVectorWrapper : public GPropWrapper {
  std::vector<T>& vect_;
public:
  GVectorWrapper(GProp& index, std::vector<T>& v) : GPropWrapper(index), vect_(v) {}
  GString stringValue() const override {return vect_.at(int(prop_->floatValue()));}
};

template <class T>
class GNumVectorWrapper : public GPropWrapper {
  std::vector<T>& vect_;
public:
  GNumVectorWrapper(GProp& index, std::vector<T>& v) : GPropWrapper(index), vect_(v) {}
  float floatValue() const override {return vect_.at(int(prop_->floatValue()));}
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template <class P, class F>
GPropRef GPropField<P,F>::operator~() {
  return GPropRef(*new GFieldWrapper<P,F>(this->prop_, this->field_));
}

template <class P, class F>
GNumPropRef GNumPropField<P,F>::operator~() {
  return GNumPropRef(*new GNumFieldWrapper<P,F>(this->prop_, this->field_));
}

class GExprWrapper : public GProp {
public:
  GExprWrapper(GExpr&);
  GExprWrapper* toExprWrapper() override {return this;}

  static GPropType& Type();
  GPropType& type() const override {return Type();}

  bool bindWith(GVarProp&);
  float floatValue() const override;
  GString stringValue() const override;

  bool isEquivalent(const GProp& other) const override {return false;}

  void onRemove(Gadget*) override;
  void removeNotifiers(void* obj) override;
  
  enum Type {TNone, TBool, TNum, TString, TText} type_{};
  gptr<GExpr> expr_;
  gptr<class GExprNotifier> notifier_;
  GVarProp* model_{};
};

}
#endif
