//
//  Multi-cursor Pong
//  Guit GUI Toolkit
//  Copyright © 2019/2020 Eric Lecolinet. All rights reserved.
//  http://www.telecom-paris.fr/~elc
//

#include <cmath>
#include <iostream>
#include <guit.hpp>
#include <gcursor.hpp>
#include <gshape.hpp>
#include <gmactouch.hpp>
using namespace std;
using namespace guit;


class GUI {
  GDim ballsize{20.f,20.f};
  GDim padsize{16.f,71.f};
  float speed = 10, angle = 0;
  int xdir = +1, ydir = +1;

  GBox       board{".board"};
  GRectShape pad1{".racket"}, pad2{".racket"};
  GOvalShape ball{".ball"};
  GTimer     timer{"20"};

public:
  GUI() {
    auto& window = Window(".gui Pong")
    << on::keydown / []{GApp::exit();}
    << on::resize  / [this]{reset();}
    << Label(".title")
    << (board << pad1 << pad2 << ball);

    // starts direct touch detection, creates as many psudo-cursors
    // as there are touchpads
    // 0 is the ID the native cursor, 1 and 2 the ID of pseudo-cursors

    auto& tm = GTouchManager::instance();
    //tm.setCaptureMouse(true);
    tm.setTimeResolution(0.05);
    auto padcount = tm.init(window);  // returns the number of touchpads
    tm.start();

    if (padcount <= 0) {
      cerr << "At least one touchpad is required" << endl;
      exit(0);
    }

    // split pad in two parts if only one pad
    if (padcount == 1) tm.getPad(0)->createCursor2(GColor::blue);

    //GCursor::cursor(0)->show(false);  // hide native cursor

    if (auto* c1 = GCursor::cursor(1)) {
      *c1 << [this,c1]{movePad(c1->wpos(), false);}; // right pad follows cursor1
      c1->show(false);   // hide cursor
    }

    if (auto* c2 = GCursor::cursor(2)) {
      *c2 << [this,c2]{movePad(c2->wpos(), true);}; // left pad follows cursor2
      c2->show(false);
    }

    timer << [this]{moveBall();};
    timer.start();

    window.startApp();   // blocking, must be called at last
  }

  
  void reset() {
    ball.resize(ballsize);
    pad1.resize(padsize);
    pad2.resize(padsize);
    ball.move((board.w() - ballsize.w)/2, (board.h() - ballsize.h)/2);
    pad1.move(15, 0);
    pad2.move(board.w() - padsize.w - 15, 0);
  }


  void movePad(GPos const& newpos, bool left) {
    float y = newpos.y();
    if (y < 0) y = 0;
    
    else if (y + pad1.h() > board.h()) y = board.h() - pad1.h();
    
    if (left) pad1.move(15, y);
    else pad2.move(board.w() - pad2.w() - 15, y);
  }


  void moveBall() {
    //static int n; cerr <<"times  #" << n++ << " "<< timer.waitingCalls() << endl;
    // waitingCalls(): includes missed calls if load is to heavy to perform all
    // of them => move faster to compensate so that the ball is always running
    // at the same speed
    float x = ball.x() + xdir * speed * timer.waitingCalls() * cos(angle);
    float y = ball.y() + ydir * speed * timer.waitingCalls() * sin(angle);
    float y1 = ball.y();
    float y2 = y1 + ball.h();
    
     if (y <= 0) {
      y = 0;
      ydir = -ydir;
    }
    else if (y >= board.h()) {
      y = board.h();
      ydir = -ydir;
    }
    if (y2 >= pad1.y() && y1 <= pad1.y()+pad1.h() && x <= pad1.x() + pad1.w()) {
      x = pad1.x() + pad1.w();
      float delta = (y + ballsize.h/2) - (pad1.y() + padsize.h/2);
      angle = delta / (padsize.h/2) * 2 * M_PI /12;
      xdir = -xdir;
    }
    else if (x <= 0) {
      x = 0;
      xdir = -xdir;
    }
    else if (y2 >= pad2.y() && y1 <= pad2.y()+pad2.h() && x + ball.w() >= pad2.x()) {
      x = pad2.x() - ball.w();
      float delta = (y + ballsize.h/2) - (pad2.y() + padsize.h/2);
      angle = -delta / (padsize.h/2) * 2 * M_PI /12;
      xdir = -xdir;
    }
    else if (x + ball.w() >= board.w()) {
      x = board.w() - ball.w();
      xdir = -xdir;
    }

    ball.move(x, y);
  }
};


int main(int argc, const char *argv[]) {
  GApp app(argc, argv);
  app
  << ".gui {icon:icon.png; align:c c; background:black;}"
   ".title {color:white; font: big; margin: 0 10; "
   "       text: First finger controls right racket, Second finger left racket - Press Esc to quit}"
   ".board {size: 550 400}"
   ".racket {background: white; size: 15 60}"
   ".ball {background: white; size: 20 20}";
  GUI gui;
  return 0;
}
